# Copyright 2018 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_utils import timeutils
import sqlalchemy

from keystone.application_credential.backends import base
from keystone.common import password_hashing
from keystone.common import sql
from keystone import exception
from keystone.i18n import _


class ApplicationCredentialModel(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'application_credential'
    attributes = [
        'internal_id',
        'id',
        'name',
        'secret_hash',
        'description',
        'user_id',
        'project_id',
        'system',
        'expires_at',
        'unrestricted',
    ]
    internal_id = sql.Column(sql.Integer, primary_key=True, nullable=False)
    id = sql.Column(sql.String(64), nullable=False)
    name = sql.Column(sql.String(255), nullable=False)
    secret_hash = sql.Column(sql.String(255), nullable=False)
    description = sql.Column(sql.Text())
    user_id = sql.Column(sql.String(64), nullable=False)
    project_id = sql.Column(sql.String(64), nullable=True)
    system = sql.Column(sql.String(64), nullable=True)
    expires_at = sql.Column(sql.DateTimeInt())
    unrestricted = sql.Column(sql.Boolean)
    __table_args__ = (
        sql.UniqueConstraint(
            'name', 'user_id', name='duplicate_app_cred_constraint'
        ),
    )

    roles = sqlalchemy.orm.relationship(
        'ApplicationCredentialRoleModel',
        backref=sqlalchemy.orm.backref('application_credential'),
        cascade='all, delete-orphan',
        cascade_backrefs=False,
    )
    access_rules = sqlalchemy.orm.relationship(
        'ApplicationCredentialAccessRuleModel',
        backref=sqlalchemy.orm.backref('application_credential'),
        cascade='all, delete-orphan',
        cascade_backrefs=False,
    )


class ApplicationCredentialRoleModel(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'application_credential_role'
    attributes = ['application_credential_id', 'role_id']
    application_credential_id = sql.Column(
        sql.Integer,
        sql.ForeignKey(
            'application_credential.internal_id', ondelete='cascade'
        ),
        primary_key=True,
        nullable=False,
    )
    role_id = sql.Column(sql.String(64), primary_key=True, nullable=False)


class AccessRuleModel(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'access_rule'
    attributes = ['external_id', 'user_id', 'service', 'path', 'method']
    id = sql.Column(sql.Integer, primary_key=True, nullable=False)
    external_id = sql.Column(sql.String(64), index=True)
    user_id = sql.Column(sql.String(64), index=True)
    service = sql.Column(sql.String(64))
    path = sql.Column(sql.String(128))
    method = sql.Column(sql.String(16))
    __table_args__ = (
        sql.UniqueConstraint(
            'user_id',
            'service',
            'path',
            'method',
            name='duplicate_access_rule_for_user_constraint',
        ),
    )
    application_credential = sqlalchemy.orm.relationship(
        'ApplicationCredentialAccessRuleModel',
        backref=sqlalchemy.orm.backref('access_rule'),
        cascade_backrefs=False,
    )


class ApplicationCredentialAccessRuleModel(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'application_credential_access_rule'
    attributes = ['application_credential_id', 'access_rule_id']
    application_credential_id = sql.Column(
        sql.Integer,
        sql.ForeignKey(
            'application_credential.internal_id', ondelete='cascade'
        ),
        primary_key=True,
        nullable=False,
    )
    access_rule_id = sql.Column(
        sql.Integer,
        sql.ForeignKey('access_rule.id'),
        primary_key=True,
        nullable=False,
    )


class ApplicationCredential(base.ApplicationCredentialDriverBase):
    def _check_secret(self, secret, app_cred_ref):
        secret_hash = app_cred_ref['secret_hash']
        return password_hashing.check_password(secret, secret_hash)

    def _check_expired(self, app_cred_ref):
        if app_cred_ref.get('expires_at'):
            return timeutils.utcnow() >= app_cred_ref['expires_at']
        return False

    def authenticate(self, application_credential_id, secret):
        msg = _('Invalid application credential ID or secret')
        try:
            app_cred_ref = self.get_application_credential(
                application_credential_id
            )
        except exception.ApplicationCredentialNotFound:
            raise AssertionError(msg)
        if not self._check_secret(secret, app_cred_ref):
            raise AssertionError(msg)
        if self._check_expired(app_cred_ref):
            raise AssertionError(msg)

    def _hash_secret(self, app_cred_ref):
        unhashed_secret = app_cred_ref.pop('secret')
        hashed_secret = password_hashing.hash_password(unhashed_secret)
        app_cred_ref['secret_hash'] = hashed_secret

    @sql.handle_conflicts(conflict_type='application_credential')
    def create_application_credential(
        self, application_credential, roles, access_rules=None
    ):
        app_cred = application_credential.copy()
        self._hash_secret(app_cred)
        with sql.session_for_write() as session:
            ref = ApplicationCredentialModel.from_dict(app_cred)
            session.add(ref)
            for role in roles:
                app_cred_role = ApplicationCredentialRoleModel()
                app_cred_role.application_credential = ref
                app_cred_role.role_id = role['id']
                session.add(app_cred_role)
            if access_rules:
                for access_rule in access_rules:
                    access_rule_ref = (
                        session.query(AccessRuleModel)
                        .filter_by(external_id=access_rule['id'])
                        .first()
                    )
                    if not access_rule_ref:
                        query = session.query(AccessRuleModel)
                        access_rule_ref = query.filter_by(
                            user_id=app_cred['user_id'],
                            service=access_rule['service'],
                            path=access_rule['path'],
                            method=access_rule['method'],
                        ).first()
                    if not access_rule_ref:
                        access_rule_ref = AccessRuleModel.from_dict(
                            {
                                k.replace('id', 'external_id'): v
                                for k, v in access_rule.items()
                            }
                        )
                        access_rule_ref['user_id'] = app_cred['user_id']
                        session.add(access_rule_ref)
                    app_cred_access_rule = (
                        ApplicationCredentialAccessRuleModel()
                    )
                    app_cred_access_rule.application_credential = ref
                    app_cred_access_rule.access_rule = access_rule_ref
                    session.add(app_cred_access_rule)
            application_credential_dict = self._to_dict(ref)
            return application_credential_dict

    def _to_dict(self, ref):
        app_cred = ref.to_dict()
        roles = [{'id': r.to_dict()['role_id']} for r in ref.roles]
        app_cred['roles'] = roles
        if ref.access_rules:
            access_rules = [
                self._access_rule_to_dict(c.access_rule)
                for c in ref.access_rules
            ]
            app_cred['access_rules'] = access_rules
        app_cred.pop('internal_id')
        return app_cred

    def _access_rule_to_dict(self, ref):
        access_rule = ref.to_dict()
        return {
            k.replace('external_id', 'id'): v
            for k, v in access_rule.items()
            if k != 'user_id' and k != 'id'
        }

    def get_application_credential(self, application_credential_id):
        with sql.session_for_read() as session:
            query = session.query(ApplicationCredentialModel).filter_by(
                id=application_credential_id
            )
            ref = query.first()
            if ref is None:
                raise exception.ApplicationCredentialNotFound(
                    application_credential_id=application_credential_id
                )
            app_cred_dict = self._to_dict(ref)
            return app_cred_dict

    def list_application_credentials_for_user(self, user_id, hints):
        with sql.session_for_read() as session:
            query = session.query(ApplicationCredentialModel)
            query = sql.filter_limit_query(
                ApplicationCredentialModel, query, hints
            )
            app_creds = query.filter_by(user_id=user_id)
            return [self._to_dict(ref) for ref in app_creds]

    @sql.handle_conflicts(conflict_type='application_credential')
    def delete_application_credential(self, application_credential_id):
        with sql.session_for_write() as session:
            query = session.query(ApplicationCredentialModel)
            app_cred_ref = query.filter_by(
                id=application_credential_id
            ).first()
            if not app_cred_ref:
                raise exception.ApplicationCredentialNotFound(
                    application_credential_id=application_credential_id
                )
            session.delete(app_cred_ref)

    def delete_application_credentials_for_user(self, user_id):
        with sql.session_for_write() as session:
            query = session.query(ApplicationCredentialModel)
            query = query.filter_by(user_id=user_id)
            query.delete()

    def delete_application_credentials_for_user_on_project(
        self, user_id, project_id
    ):
        with sql.session_for_write() as session:
            query = session.query(ApplicationCredentialModel)
            query = query.filter_by(user_id=user_id)
            query = query.filter_by(project_id=project_id)
            query.delete()

    def get_access_rule(self, access_rule_id):
        with sql.session_for_read() as session:
            query = session.query(AccessRuleModel).filter_by(
                external_id=access_rule_id
            )
            ref = query.first()
            if not ref:
                raise exception.AccessRuleNotFound(
                    access_rule_id=access_rule_id
                )
            access_rule = self._access_rule_to_dict(ref)
            return access_rule

    def list_access_rules_for_user(self, user_id, hints):
        with sql.session_for_read() as session:
            query = session.query(AccessRuleModel).filter_by(user_id=user_id)
            refs = sql.filter_limit_query(AccessRuleModel, query, hints)
            return [self._access_rule_to_dict(ref) for ref in refs]

    def delete_access_rule(self, access_rule_id):
        try:
            with sql.session_for_write() as session:
                query = session.query(AccessRuleModel)
                ref = query.filter_by(external_id=access_rule_id).first()
                if not ref:
                    raise exception.AccessRuleNotFound(
                        access_rule_id=access_rule_id
                    )
                session.delete(ref)
        except AssertionError:
            raise exception.ForbiddenNotSecurity(
                "May not delete access rule in use"
            )

    def delete_access_rules_for_user(self, user_id):
        with sql.session_for_write() as session:
            query = session.query(AccessRuleModel).filter_by(user_id=user_id)
            query.delete()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The domain config API is now aware of system scope and default roles."
)

deprecated_get_domain_config = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_domain_config',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_get_domain_config_default = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_domain_config_default',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_create_domain_config = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_domain_config',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_update_domain_config = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_domain_config',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_delete_domain_config = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_domain_config',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)


domain_config_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_domain_config',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create domain configuration.',
        operations=[
            {'path': '/v3/domains/{domain_id}/config', 'method': 'PUT'}
        ],
        deprecated_rule=deprecated_create_domain_config,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_domain_config',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description=(
            'Get the entire domain configuration for a domain, an '
            'option group within a domain, or a specific '
            'configuration option within a group for a domain.'
        ),
        operations=[
            {'path': '/v3/domains/{domain_id}/config', 'method': 'GET'},
            {'path': '/v3/domains/{domain_id}/config', 'method': 'HEAD'},
            {
                'path': '/v3/domains/{domain_id}/config/{group}',
                'method': 'GET',
            },
            {
                'path': '/v3/domains/{domain_id}/config/{group}',
                'method': 'HEAD',
            },
            {
                'path': '/v3/domains/{domain_id}/config/{group}/{option}',
                'method': 'GET',
            },
            {
                'path': '/v3/domains/{domain_id}/config/{group}/{option}',
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_get_domain_config,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_security_compliance_domain_config',
        check_str='',
        # This should be accessible to anyone with a valid token, regardless of
        # system-scope or project-scope.
        scope_types=['system', 'domain', 'project'],
        description=(
            'Get security compliance domain configuration for '
            'either a domain or a specific option in a domain.'
        ),
        operations=[
            {
                'path': '/v3/domains/{domain_id}/config/security_compliance',
                'method': 'GET',
            },
            {
                'path': '/v3/domains/{domain_id}/config/security_compliance',
                'method': 'HEAD',
            },
            {
                'path': (
                    '/v3/domains/{domain_id}/config/'
                    'security_compliance/{option}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/domains/{domain_id}/config/'
                    'security_compliance/{option}'
                ),
                'method': 'HEAD',
            },
        ],
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_domain_config',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description=(
            'Update domain configuration for either a domain, '
            'specific group or a specific option in a group.'
        ),
        operations=[
            {'path': '/v3/domains/{domain_id}/config', 'method': 'PATCH'},
            {
                'path': '/v3/domains/{domain_id}/config/{group}',
                'method': 'PATCH',
            },
            {
                'path': '/v3/domains/{domain_id}/config/{group}/{option}',
                'method': 'PATCH',
            },
        ],
        deprecated_rule=deprecated_update_domain_config,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_domain_config',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description=(
            'Delete domain configuration for either a domain, '
            'specific group or a specific option in a group.'
        ),
        operations=[
            {'path': '/v3/domains/{domain_id}/config', 'method': 'DELETE'},
            {
                'path': '/v3/domains/{domain_id}/config/{group}',
                'method': 'DELETE',
            },
            {
                'path': '/v3/domains/{domain_id}/config/{group}/{option}',
                'method': 'DELETE',
            },
        ],
        deprecated_rule=deprecated_delete_domain_config,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_domain_config_default',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description=(
            'Get domain configuration default for either a domain, '
            'specific group or a specific option in a group.'
        ),
        operations=[
            {'path': '/v3/domains/config/default', 'method': 'GET'},
            {'path': '/v3/domains/config/default', 'method': 'HEAD'},
            {'path': '/v3/domains/config/{group}/default', 'method': 'GET'},
            {'path': '/v3/domains/config/{group}/default', 'method': 'HEAD'},
            {
                'path': '/v3/domains/config/{group}/{option}/default',
                'method': 'GET',
            },
            {
                'path': '/v3/domains/config/{group}/{option}/default',
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_get_domain_config_default,
    ),
]


def list_rules():
    return domain_config_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

# NOTE(lbragstad): Both endpoints and services are system-level resources.
# System-scoped tokens should be required to manage policy associations to
# existing system-level resources.

DEPRECATED_REASON = (
    "The policy association API is now aware of system scope and default "
    "roles."
)

deprecated_check_policy_assoc_for_endpoint = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_policy_association_for_endpoint',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_check_policy_assoc_for_service = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_policy_association_for_service',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_check_policy_assoc_for_region_and_service = policy.DeprecatedRule(
    name=base.IDENTITY % 'check_policy_association_for_region_and_service',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_get_policy_for_endpoint = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_policy_for_endpoint',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_list_endpoints_for_policy = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoints_for_policy',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_create_policy_assoc_for_endpoint = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_policy_association_for_endpoint',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_delete_policy_assoc_for_endpoint = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_policy_association_for_endpoint',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_create_policy_assoc_for_service = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_policy_association_for_service',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_delete_policy_assoc_for_service = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_policy_association_for_service',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_create_policy_assoc_for_region_and_service = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_policy_association_for_region_and_service',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_delete_policy_assoc_for_region_and_service = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_policy_association_for_region_and_service',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)


policy_association_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_policy_association_for_endpoint',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Associate a policy to a specific endpoint.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'PUT',
            }
        ],
        deprecated_rule=deprecated_create_policy_assoc_for_endpoint,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_policy_association_for_endpoint',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Check policy association for endpoint.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_check_policy_assoc_for_endpoint,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_policy_association_for_endpoint',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete policy association for endpoint.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'endpoints/{endpoint_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_delete_policy_assoc_for_endpoint,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_policy_association_for_service',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Associate a policy to a specific service.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}'
                ),
                'method': 'PUT',
            }
        ],
        deprecated_rule=deprecated_create_policy_assoc_for_service,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_policy_association_for_service',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Check policy association for service.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_check_policy_assoc_for_service,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_policy_association_for_service',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete policy association for service.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_delete_policy_assoc_for_service,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY
        % ('create_policy_association_for_region_and_service'),
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description=(
            'Associate a policy to a specific region and service '
            'combination.'
        ),
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}/regions/{region_id}'
                ),
                'method': 'PUT',
            }
        ],
        deprecated_rule=deprecated_create_policy_assoc_for_region_and_service,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'check_policy_association_for_region_and_service',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Check policy association for region and service.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}/regions/{region_id}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}/regions/{region_id}'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_check_policy_assoc_for_region_and_service,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY
        % ('delete_policy_association_for_region_and_service'),
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete policy association for region and service.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/'
                    'services/{service_id}/regions/{region_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_delete_policy_assoc_for_region_and_service,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_policy_for_endpoint',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Get policy for endpoint.',
        operations=[
            {
                'path': (
                    '/v3/endpoints/{endpoint_id}/OS-ENDPOINT-POLICY/policy'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/endpoints/{endpoint_id}/OS-ENDPOINT-POLICY/policy'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_get_policy_for_endpoint,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoints_for_policy',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List endpoints for policy.',
        operations=[
            {
                'path': (
                    '/v3/policies/{policy_id}/OS-ENDPOINT-POLICY/endpoints'
                ),
                'method': 'GET',
            }
        ],
        deprecated_rule=deprecated_list_endpoints_for_policy,
    ),
]


def list_rules():
    return policy_association_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone.common import provider_api
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit.identity.shadow_users import test_backend
from keystone.tests.unit.identity.shadow_users import test_core
from keystone.tests.unit.ksfixtures import database

PROVIDERS = provider_api.ProviderAPIs


class ShadowUsersTests(
    unit.TestCase,
    test_backend.ShadowUsersBackendTests,
    test_core.ShadowUsersCoreTests,
):
    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.load_backends()
        PROVIDERS.resource_api.create_domain(
            default_fixtures.ROOT_DOMAIN['id'], default_fixtures.ROOT_DOMAIN
        )
        self.idp = {
            'id': uuid.uuid4().hex,
            'enabled': True,
            'description': uuid.uuid4().hex,
        }
        self.mapping = {'id': uuid.uuid4().hex}
        self.protocol = {
            'id': uuid.uuid4().hex,
            'idp_id': self.idp['id'],
            'mapping_id': self.mapping['id'],
        }
        self.federated_user = {
            'idp_id': self.idp['id'],
            'protocol_id': self.protocol['id'],
            'unique_id': uuid.uuid4().hex,
            'display_name': uuid.uuid4().hex,
        }
        self.email = uuid.uuid4().hex
        PROVIDERS.federation_api.create_idp(self.idp['id'], self.idp)
        PROVIDERS.federation_api.create_mapping(
            self.mapping['id'], self.mapping
        )
        PROVIDERS.federation_api.create_protocol(
            self.idp['id'], self.protocol['id'], self.protocol
        )
        self.domain_id = PROVIDERS.federation_api.get_idp(self.idp['id'])[
            'domain_id'
        ]


class TestUserWithFederatedUser(ShadowUsersTests):
    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self.load_backends()

    def assertFederatedDictsEqual(self, fed_dict, fed_object):
        self.assertEqual(fed_dict['idp_id'], fed_object['idp_id'])
        self.assertEqual(
            fed_dict['protocol_id'], fed_object['protocols'][0]['protocol_id']
        )
        self.assertEqual(
            fed_dict['unique_id'], fed_object['protocols'][0]['unique_id']
        )

    def test_get_user_when_user_has_federated_object(self):
        fed_dict = unit.new_federated_user_ref(
            idp_id=self.idp['id'], protocol_id=self.protocol['id']
        )
        user = self.shadow_users_api.create_federated_user(
            self.domain_id, fed_dict
        )

        # test that the user returns a federated object and that there is only
        # one returned
        user_ref = self.identity_api.get_user(user['id'])
        self.assertIn('federated', user_ref)
        self.assertEqual(1, len(user_ref['federated']))

        self.assertFederatedDictsEqual(fed_dict, user_ref['federated'][0])

    def test_create_user_with_invalid_idp_and_protocol_fails(self):
        baduser = unit.new_user_ref(domain_id=self.domain_id)
        baduser['federated'] = [
            {
                'idp_id': 'fakeidp',
                'protocols': [
                    {'protocol_id': 'nonexistent', 'unique_id': 'unknown'}
                ],
            }
        ]
        # Check validation works by throwing a federated object with
        # invalid idp_id, protocol_id inside the user passed to create_user.
        self.assertRaises(
            exception.ValidationError, self.identity_api.create_user, baduser
        )

        baduser['federated'][0]['idp_id'] = self.idp['id']
        self.assertRaises(
            exception.ValidationError, self.identity_api.create_user, baduser
        )

    def test_create_user_with_federated_attributes(self):
        # Create the schema of a federated attribute being passed in with a
        # user.
        user = unit.new_user_ref(domain_id=self.domain_id)
        unique_id = uuid.uuid4().hex
        user['federated'] = [
            {
                'idp_id': self.idp['id'],
                'protocols': [
                    {
                        'protocol_id': self.protocol['id'],
                        'unique_id': unique_id,
                    }
                ],
            }
        ]

        # Test that there are no current federated_users that match our users
        # federated object and create the user
        self.assertRaises(
            exception.UserNotFound,
            self.shadow_users_api.get_federated_user,
            self.idp['id'],
            self.protocol['id'],
            unique_id,
        )

        ref = self.identity_api.create_user(user)

        # Test that the user and federated object now exists
        self.assertEqual(user['name'], ref['name'])
        self.assertEqual(user['federated'], ref['federated'])
        fed_user = self.shadow_users_api.get_federated_user(
            self.idp['id'], self.protocol['id'], unique_id
        )
        self.assertIsNotNone(fed_user)

    def test_update_user_with_invalid_idp_and_protocol_fails(self):
        baduser = unit.new_user_ref(domain_id=self.domain_id)
        baduser['federated'] = [
            {
                'idp_id': 'fakeidp',
                'protocols': [
                    {'protocol_id': 'nonexistent', 'unique_id': 'unknown'}
                ],
            }
        ]
        # Check validation works by throwing a federated object with
        # invalid idp_id, protocol_id inside the user passed to create_user.
        self.assertRaises(
            exception.ValidationError, self.identity_api.create_user, baduser
        )

        baduser['federated'][0]['idp_id'] = self.idp['id']
        self.assertRaises(
            exception.ValidationError, self.identity_api.create_user, baduser
        )

    def test_update_user_with_federated_attributes(self):
        user = self.shadow_users_api.create_federated_user(
            self.domain_id, self.federated_user
        )
        user = self.identity_api.get_user(user['id'])

        # Test that update user can return a federated object with the user as
        # a response if the user has any
        user = self.identity_api.update_user(user['id'], user)
        self.assertFederatedDictsEqual(
            self.federated_user, user['federated'][0]
        )

        # Test that update user can replace a users federated objects if added
        # in the request and that its response is that new federated objects
        new_fed = [
            {
                'idp_id': self.idp['id'],
                'protocols': [
                    {
                        'protocol_id': self.protocol['id'],
                        'unique_id': uuid.uuid4().hex,
                    }
                ],
            }
        ]
        user['federated'] = new_fed
        user = self.identity_api.update_user(user['id'], user)
        self.assertTrue('federated' in user)
        self.assertEqual(len(user['federated']), 1)
        self.assertEqual(user['federated'][0], new_fed[0])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings

import fixtures
from sqlalchemy import exc as sqla_exc


class WarningsFixture(fixtures.Fixture):
    """Filters out warnings during test runs."""

    def setUp(self):
        super().setUp()

        self._original_warning_filters = warnings.filters[:]

        # NOTE(stephenfin): Make deprecation warnings only happen once.
        # Otherwise this gets kind of crazy given the way that upstream python
        # libs use this.
        warnings.simplefilter('once', DeprecationWarning)

        warnings.filterwarnings(
            'error', module='keystone', category=DeprecationWarning
        )

        warnings.filterwarnings(
            'ignore',
            message=(
                'Policy enforcement is depending on the value of '
                '(token|group_ids). '
                'This key is deprecated. Please update your policy '
                'file to use the standard policy values.'
            ),
        )

        # NOTE(stephenfin): Ignore scope check UserWarnings from oslo.policy.
        warnings.filterwarnings(
            'ignore',
            message="Policy .* failed scope check",
            category=UserWarning,
        )

        # TODO(stephenfin): This will be fixed once we drop sqlalchemy-migrate
        warnings.filterwarnings(
            'ignore',
            category=DeprecationWarning,
            message=r"Using function/method 'db_version\(\)' is deprecated",
        )

        warnings.filterwarnings(
            'error', module='keystone', category=sqla_exc.SAWarning
        )

        warnings.filterwarnings(
            'ignore', category=sqla_exc.SADeprecationWarning
        )

        # Enable deprecation warnings for keystone itself to capture upcoming
        # SQLALchemy changes

        warnings.filterwarnings(
            'error', module='keystone', category=sqla_exc.SADeprecationWarning
        )

        self.addCleanup(self._reset_warning_filters)

    def _reset_warning_filters(self):
        warnings.filters[:] = self._original_warning_filters



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import http.client
import uuid

from testtools import matchers

from keystone.common import provider_api
from keystone.tests import unit
from keystone.tests.unit import test_v3

PROVIDERS = provider_api.ProviderAPIs


class EndpointFilterTestCase(test_v3.RestfulTestCase):
    def setUp(self):
        super().setUp()
        self.default_request_url = (
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{self.endpoint_id}'
        )


class EndpointFilterCRUDTestCase(EndpointFilterTestCase):
    def test_create_endpoint_project_association(self):
        """PUT /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Valid endpoint and project id test case.

        """
        self.put(self.default_request_url)

    def test_create_endpoint_project_association_with_invalid_project(self):
        """PUT OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid project id test case.

        """
        self.put(
            f'/OS-EP-FILTER/projects/{uuid.uuid4().hex}'
            f'/endpoints/{self.endpoint_id}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_create_endpoint_project_association_with_invalid_endpoint(self):
        """PUT /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid endpoint id test case.

        """
        self.put(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{uuid.uuid4().hex}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_create_endpoint_project_association_with_unexpected_body(self):
        """PUT /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Unexpected body in request. The body should be ignored.

        """
        self.put(
            self.default_request_url,
            body={'project_id': self.default_domain_project_id},
        )

    def test_check_endpoint_project_association(self):
        """HEAD /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Valid project and endpoint id test case.

        """
        self.put(self.default_request_url)
        self.head(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{self.endpoint_id}',
            expected_status=http.client.NO_CONTENT,
        )

    def test_check_endpoint_project_association_with_invalid_project(self):
        """HEAD /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid project id test case.

        """
        self.put(self.default_request_url)
        self.head(
            f'/OS-EP-FILTER/projects/{uuid.uuid4().hex}'
            f'/endpoints/{self.endpoint_id}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_check_endpoint_project_association_with_invalid_endpoint(self):
        """HEAD /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid endpoint id test case.

        """
        self.put(self.default_request_url)
        self.head(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{uuid.uuid4().hex}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_get_endpoint_project_association(self):
        """GET /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Valid project and endpoint id test case.

        """
        self.put(self.default_request_url)
        self.get(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{self.endpoint_id}',
            expected_status=http.client.NO_CONTENT,
        )

    def test_get_endpoint_project_association_with_invalid_project(self):
        """GET /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid project id test case.

        """
        self.put(self.default_request_url)
        self.get(
            f'/OS-EP-FILTER/projects/{uuid.uuid4().hex}'
            f'/endpoints/{self.endpoint_id}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_get_endpoint_project_association_with_invalid_endpoint(self):
        """GET /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid endpoint id test case.

        """
        self.put(self.default_request_url)
        self.get(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{uuid.uuid4().hex}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_list_endpoints_associated_with_valid_project(self):
        """GET & HEAD /OS-EP-FILTER/projects/{project_id}/endpoints.

        Valid project and endpoint id test case.

        """
        self.put(self.default_request_url)
        resource_url = f'/OS-EP-FILTER/projects/{self.default_domain_project_id}/endpoints'
        r = self.get(resource_url)
        self.assertValidEndpointListResponse(
            r, self.endpoint, resource_url=resource_url
        )
        self.head(resource_url, expected_status=http.client.OK)

    def test_list_endpoints_associated_with_invalid_project(self):
        """GET & HEAD /OS-EP-FILTER/projects/{project_id}/endpoints.

        Invalid project id test case.

        """
        self.put(self.default_request_url)
        url = f'/OS-EP-FILTER/projects/{uuid.uuid4().hex}/endpoints'
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_list_projects_associated_with_endpoint(self):
        """GET & HEAD /OS-EP-FILTER/endpoints/{endpoint_id}/projects.

        Valid endpoint-project association test case.

        """
        self.put(self.default_request_url)
        resource_url = f'/OS-EP-FILTER/endpoints/{self.endpoint_id}/projects'
        r = self.get(resource_url, expected_status=http.client.OK)
        self.assertValidProjectListResponse(
            r, self.default_domain_project, resource_url=resource_url
        )
        self.head(resource_url, expected_status=http.client.OK)

    def test_list_projects_with_no_endpoint_project_association(self):
        """GET & HEAD /OS-EP-FILTER/endpoints/{endpoint_id}/projects.

        Valid endpoint id but no endpoint-project associations test case.

        """
        url = f'/OS-EP-FILTER/endpoints/{self.endpoint_id}/projects'
        r = self.get(url, expected_status=http.client.OK)
        self.assertValidProjectListResponse(r, expected_length=0)
        self.head(url, expected_status=http.client.OK)

    def test_list_projects_associated_with_invalid_endpoint(self):
        """GET & HEAD /OS-EP-FILTER/endpoints/{endpoint_id}/projects.

        Invalid endpoint id test case.

        """
        url = f'/OS-EP-FILTER/endpoints/{uuid.uuid4().hex}/projects'
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_remove_endpoint_project_association(self):
        """DELETE /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Valid project id and endpoint id test case.

        """
        self.put(self.default_request_url)
        self.delete(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{self.endpoint_id}'
        )

    def test_remove_endpoint_project_association_with_invalid_project(self):
        """DELETE /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid project id test case.

        """
        self.put(self.default_request_url)
        self.delete(
            f'/OS-EP-FILTER/projects/{uuid.uuid4().hex}'
            f'/endpoints/{self.endpoint_id}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_remove_endpoint_project_association_with_invalid_endpoint(self):
        """DELETE /OS-EP-FILTER/projects/{project_id}/endpoints/{endpoint_id}.

        Invalid endpoint id test case.

        """
        self.put(self.default_request_url)
        self.delete(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{uuid.uuid4().hex}',
            expected_status=http.client.NOT_FOUND,
        )

    def test_endpoint_project_association_cleanup_when_project_deleted(self):
        self.put(self.default_request_url)
        association_url = (
            f'/OS-EP-FILTER/endpoints/{self.endpoint_id}/projects'
        )
        r = self.get(association_url)
        self.assertValidProjectListResponse(r, expected_length=1)

        self.delete(f'/projects/{self.default_domain_project_id}')

        r = self.get(association_url)
        self.assertValidProjectListResponse(r, expected_length=0)

    def test_endpoint_project_association_cleanup_when_endpoint_deleted(self):
        self.put(self.default_request_url)
        association_url = f'/OS-EP-FILTER/projects/{self.default_domain_project_id}/endpoints'
        r = self.get(association_url)
        self.assertValidEndpointListResponse(r, expected_length=1)

        self.delete(f'/endpoints/{self.endpoint_id}')

        r = self.get(association_url)
        self.assertValidEndpointListResponse(r, expected_length=0)

    @unit.skip_if_cache_disabled('catalog')
    def test_create_endpoint_project_association_invalidates_cache(self):
        # NOTE(davechen): create another endpoint which will be added to
        # default project, this should be done at first since
        # `create_endpoint` will also invalidate cache.
        endpoint_id2 = uuid.uuid4().hex
        endpoint2 = unit.new_endpoint_ref(
            service_id=self.service_id,
            region_id=self.region_id,
            interface='public',
            id=endpoint_id2,
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_id2, endpoint2.copy())

        # create endpoint project association.
        self.put(self.default_request_url)

        # should get back only one endpoint that was just created.
        user_id = uuid.uuid4().hex
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        # there is only one endpoints associated with the default project.
        self.assertEqual(1, len(catalog[0]['endpoints']))
        self.assertEqual(self.endpoint_id, catalog[0]['endpoints'][0]['id'])

        # add the second endpoint to default project, bypassing
        # catalog_api API manager.
        PROVIDERS.catalog_api.driver.add_endpoint_to_project(
            endpoint_id2, self.default_domain_project_id
        )

        # but, we can just get back one endpoint from the cache, since the
        # catalog is pulled out from cache and its haven't been invalidated.
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertEqual(1, len(catalog[0]['endpoints']))

        # remove the endpoint2 from the default project, and add it again via
        # catalog_api API manager.
        PROVIDERS.catalog_api.driver.remove_endpoint_from_project(
            endpoint_id2, self.default_domain_project_id
        )

        # add second endpoint to default project, this can be done by calling
        # the catalog_api API manager directly but call the REST API
        # instead for consistency.
        self.put(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{endpoint_id2}'
        )

        # should get back two endpoints since the cache has been
        # invalidated when the second endpoint was added to default project.
        catalog = self.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertEqual(2, len(catalog[0]['endpoints']))

        ep_id_list = [
            catalog[0]['endpoints'][0]['id'],
            catalog[0]['endpoints'][1]['id'],
        ]
        self.assertCountEqual([self.endpoint_id, endpoint_id2], ep_id_list)

    @unit.skip_if_cache_disabled('catalog')
    def test_remove_endpoint_from_project_invalidates_cache(self):
        endpoint_id2 = uuid.uuid4().hex
        endpoint2 = unit.new_endpoint_ref(
            service_id=self.service_id,
            region_id=self.region_id,
            interface='public',
            id=endpoint_id2,
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_id2, endpoint2.copy())
        # create endpoint project association.
        self.put(self.default_request_url)

        # add second endpoint to default project.
        self.put(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{endpoint_id2}'
        )

        # should get back only one endpoint that was just created.
        user_id = uuid.uuid4().hex
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        # there are two endpoints associated with the default project.
        ep_id_list = [
            catalog[0]['endpoints'][0]['id'],
            catalog[0]['endpoints'][1]['id'],
        ]
        self.assertEqual(2, len(catalog[0]['endpoints']))
        self.assertCountEqual([self.endpoint_id, endpoint_id2], ep_id_list)

        # remove the endpoint2 from the default project, bypassing
        # catalog_api API manager.
        PROVIDERS.catalog_api.driver.remove_endpoint_from_project(
            endpoint_id2, self.default_domain_project_id
        )

        # but, we can just still get back two endpoints from the cache,
        # since the catalog is pulled out from cache and its haven't
        # been invalidated.
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertEqual(2, len(catalog[0]['endpoints']))

        # add back the endpoint2 to the default project, and remove it by
        # catalog_api API manage.
        PROVIDERS.catalog_api.driver.add_endpoint_to_project(
            endpoint_id2, self.default_domain_project_id
        )

        # remove the endpoint2 from the default project, this can be done
        # by calling the catalog_api API manager directly but call
        # the REST API instead for consistency.
        self.delete(
            f'/OS-EP-FILTER/projects/{self.default_domain_project_id}'
            f'/endpoints/{endpoint_id2}'
        )

        # should only get back one endpoint since the cache has been
        # invalidated after the endpoint project association was removed.
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertEqual(1, len(catalog[0]['endpoints']))
        self.assertEqual(self.endpoint_id, catalog[0]['endpoints'][0]['id'])


class EndpointFilterTokenRequestTestCase(EndpointFilterTestCase):
    def test_project_scoped_token_using_endpoint_filter(self):
        """Verify endpoints from project scoped token filtered."""
        # create a project to work with
        ref = unit.new_project_ref(domain_id=self.domain_id)
        r = self.post('/projects', body={'project': ref})
        project = self.assertValidProjectResponse(r, ref)

        # grant the user a role on the project
        self.put(
            '/projects/{project_id}/users/{user_id}/roles/{role_id}'.format(
                user_id=self.user['id'],
                project_id=project['id'],
                role_id=self.role['id'],
            )
        )

        # set the user's preferred project
        body = {'user': {'default_project_id': project['id']}}
        r = self.patch(
            '/users/{user_id}'.format(user_id=self.user['id']), body=body
        )
        self.assertValidUserResponse(r)

        # add one endpoint to the project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=project['id'], endpoint_id=self.endpoint_id
            )
        )

        # attempt to authenticate without requesting a project
        auth_data = self.build_authentication_request(
            user_id=self.user['id'], password=self.user['password']
        )
        r = self.post('/auth/tokens', body=auth_data)
        self.assertValidProjectScopedTokenResponse(
            r, require_catalog=True, endpoint_filter=True, ep_filter_assoc=1
        )
        self.assertEqual(project['id'], r.result['token']['project']['id'])

    def test_default_scoped_token_using_endpoint_filter(self):
        """Verify endpoints from default scoped token filtered."""
        # add one endpoint to default project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=self.endpoint_id
            )
        )

        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )
        r = self.post('/auth/tokens', body=auth_data)
        self.assertValidProjectScopedTokenResponse(
            r, require_catalog=True, endpoint_filter=True, ep_filter_assoc=1
        )
        self.assertEqual(
            self.project['id'], r.result['token']['project']['id']
        )

        # Ensure name of the service exists
        self.assertIn('name', r.result['token']['catalog'][0])

        # region and region_id should be the same in endpoints
        endpoint = r.result['token']['catalog'][0]['endpoints'][0]
        self.assertIn('region', endpoint)
        self.assertIn('region_id', endpoint)
        self.assertEqual(endpoint['region'], endpoint['region_id'])

    def test_scoped_token_with_no_catalog_using_endpoint_filter(self):
        """Verify endpoint filter does not affect no catalog."""
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=self.endpoint_id
            )
        )

        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )
        r = self.post('/auth/tokens?nocatalog', body=auth_data)
        self.assertValidProjectScopedTokenResponse(r, require_catalog=False)
        self.assertEqual(
            self.project['id'], r.result['token']['project']['id']
        )

    def test_invalid_endpoint_project_association(self):
        """Verify an invalid endpoint-project association is handled."""
        # add first endpoint to default project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=self.endpoint_id
            )
        )

        # create a second temporary endpoint
        endpoint_id2 = uuid.uuid4().hex
        endpoint2 = unit.new_endpoint_ref(
            service_id=self.service_id,
            region_id=self.region_id,
            interface='public',
            id=endpoint_id2,
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_id2, endpoint2.copy())

        # add second endpoint to default project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=endpoint_id2
            )
        )

        # remove the temporary reference
        # this will create inconsistency in the endpoint filter table
        # which is fixed during the catalog creation for token request
        PROVIDERS.catalog_api.delete_endpoint(endpoint_id2)

        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )
        r = self.post('/auth/tokens', body=auth_data)
        self.assertValidProjectScopedTokenResponse(
            r, require_catalog=True, endpoint_filter=True, ep_filter_assoc=1
        )
        self.assertEqual(
            self.project['id'], r.result['token']['project']['id']
        )

    def test_disabled_endpoint(self):
        """Test that a disabled endpoint is handled."""
        # Add an enabled endpoint to the default project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=self.endpoint_id
            )
        )

        # Add a disabled endpoint to the default project.

        # Create a disabled endpoint that's like the enabled one.
        disabled_endpoint_ref = copy.copy(self.endpoint)
        disabled_endpoint_id = uuid.uuid4().hex
        disabled_endpoint_ref.update(
            {
                'id': disabled_endpoint_id,
                'enabled': False,
                'interface': 'internal',
            }
        )
        PROVIDERS.catalog_api.create_endpoint(
            disabled_endpoint_id, disabled_endpoint_ref
        )

        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=disabled_endpoint_id
            )
        )

        # Authenticate to get token with catalog
        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )
        r = self.post('/auth/tokens', body=auth_data)

        endpoints = r.result['token']['catalog'][0]['endpoints']
        endpoint_ids = [ep['id'] for ep in endpoints]
        self.assertEqual([self.endpoint_id], endpoint_ids)

    def test_multiple_endpoint_project_associations(self):
        def _create_an_endpoint():
            endpoint_ref = unit.new_endpoint_ref(
                service_id=self.service_id,
                interface='public',
                region_id=self.region_id,
            )
            r = self.post('/endpoints', body={'endpoint': endpoint_ref})
            return r.result['endpoint']['id']

        # create three endpoints
        endpoint_id1 = _create_an_endpoint()
        endpoint_id2 = _create_an_endpoint()
        _create_an_endpoint()

        # only associate two endpoints with project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=endpoint_id1
            )
        )
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=endpoint_id2
            )
        )

        # there should be only two endpoints in token catalog
        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )
        r = self.post('/auth/tokens', body=auth_data)
        self.assertValidProjectScopedTokenResponse(
            r, require_catalog=True, endpoint_filter=True, ep_filter_assoc=2
        )

    def test_get_auth_catalog_using_endpoint_filter(self):
        # add one endpoint to default project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=self.endpoint_id
            )
        )

        auth_data = self.build_authentication_request(
            user_id=self.user['id'],
            password=self.user['password'],
            project_id=self.project['id'],
        )
        token_data = self.post('/auth/tokens', body=auth_data)
        self.assertValidProjectScopedTokenResponse(
            token_data,
            require_catalog=True,
            endpoint_filter=True,
            ep_filter_assoc=1,
        )

        auth_catalog = self.get(
            '/auth/catalog', token=token_data.headers['X-Subject-Token']
        )
        self.assertEqual(
            token_data.result['token']['catalog'],
            auth_catalog.result['catalog'],
        )


class JsonHomeTests(EndpointFilterTestCase, test_v3.JsonHomeTestMixin):
    JSON_HOME_DATA = {
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-EP-FILTER/'
        '1.0/rel/endpoint_projects': {
            'href-template': '/OS-EP-FILTER/endpoints/{endpoint_id}/projects',
            'href-vars': {
                'endpoint_id': 'https://docs.openstack.org/api/openstack-identity/3/param/'
                'endpoint_id'
            },
        },
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-EP-FILTER/'
        '1.0/rel/endpoint_groups': {'href': '/OS-EP-FILTER/endpoint_groups'},
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-EP-FILTER/'
        '1.0/rel/endpoint_group': {
            'href-template': '/OS-EP-FILTER/endpoint_groups/'
            '{endpoint_group_id}',
            'href-vars': {
                'endpoint_group_id': 'https://docs.openstack.org/api/openstack-identity/3/'
                'ext/OS-EP-FILTER/1.0/param/endpoint_group_id'
            },
        },
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-EP-FILTER/'
        '1.0/rel/endpoint_group_to_project_association': {
            'href-template': '/OS-EP-FILTER/endpoint_groups/'
            '{endpoint_group_id}/projects/{project_id}',
            'href-vars': {
                'project_id': 'https://docs.openstack.org/api/openstack-identity/3/param/'
                'project_id',
                'endpoint_group_id': 'https://docs.openstack.org/api/openstack-identity/3/'
                'ext/OS-EP-FILTER/1.0/param/endpoint_group_id',
            },
        },
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-EP-FILTER/'
        '1.0/rel/projects_associated_with_endpoint_group': {
            'href-template': '/OS-EP-FILTER/endpoint_groups/'
            '{endpoint_group_id}/projects',
            'href-vars': {
                'endpoint_group_id': 'https://docs.openstack.org/api/openstack-identity/3/'
                'ext/OS-EP-FILTER/1.0/param/endpoint_group_id'
            },
        },
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-EP-FILTER/'
        '1.0/rel/endpoints_in_endpoint_group': {
            'href-template': '/OS-EP-FILTER/endpoint_groups/'
            '{endpoint_group_id}/endpoints',
            'href-vars': {
                'endpoint_group_id': 'https://docs.openstack.org/api/openstack-identity/3/'
                'ext/OS-EP-FILTER/1.0/param/endpoint_group_id'
            },
        },
        'https://docs.openstack.org/api/openstack-identity/3/ext/OS-EP-FILTER/'
        '1.0/rel/project_endpoint_groups': {
            'href-template': '/OS-EP-FILTER/projects/{project_id}/'
            'endpoint_groups',
            'href-vars': {
                'project_id': 'https://docs.openstack.org/api/openstack-identity/3/param/'
                'project_id'
            },
        },
    }


class EndpointGroupCRUDTestCase(EndpointFilterTestCase):
    DEFAULT_ENDPOINT_GROUP_BODY = {
        'endpoint_group': {
            'description': 'endpoint group description',
            'filters': {'interface': 'admin'},
            'name': 'endpoint_group_name',
        }
    }

    DEFAULT_ENDPOINT_GROUP_URL = '/OS-EP-FILTER/endpoint_groups'

    def test_create_endpoint_group(self):
        """POST /OS-EP-FILTER/endpoint_groups.

        Valid endpoint group test case.

        """
        r = self.post(
            self.DEFAULT_ENDPOINT_GROUP_URL,
            body=self.DEFAULT_ENDPOINT_GROUP_BODY,
        )
        expected_filters = self.DEFAULT_ENDPOINT_GROUP_BODY['endpoint_group'][
            'filters'
        ]
        expected_name = self.DEFAULT_ENDPOINT_GROUP_BODY['endpoint_group'][
            'name'
        ]
        self.assertEqual(
            expected_filters, r.result['endpoint_group']['filters']
        )
        self.assertEqual(expected_name, r.result['endpoint_group']['name'])
        self.assertThat(
            r.result['endpoint_group']['links']['self'],
            matchers.EndsWith(
                '/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'.format(
                    endpoint_group_id=r.result['endpoint_group']['id']
                )
            ),
        )

    def test_create_invalid_endpoint_group(self):
        """POST /OS-EP-FILTER/endpoint_groups.

        Invalid endpoint group creation test case.

        """
        invalid_body = copy.deepcopy(self.DEFAULT_ENDPOINT_GROUP_BODY)
        invalid_body['endpoint_group']['filters'] = {'foobar': 'admin'}
        self.post(
            self.DEFAULT_ENDPOINT_GROUP_URL,
            body=invalid_body,
            expected_status=http.client.BAD_REQUEST,
        )

    def test_get_endpoint_group(self):
        """GET /OS-EP-FILTER/endpoint_groups/{endpoint_group}.

        Valid endpoint group test case.

        """
        # create an endpoint group to work with
        response = self.post(
            self.DEFAULT_ENDPOINT_GROUP_URL,
            body=self.DEFAULT_ENDPOINT_GROUP_BODY,
        )
        endpoint_group_id = response.result['endpoint_group']['id']
        endpoint_group_filters = response.result['endpoint_group']['filters']
        endpoint_group_name = response.result['endpoint_group']['name']
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.get(url)
        self.assertEqual(
            endpoint_group_id, response.result['endpoint_group']['id']
        )
        self.assertEqual(
            endpoint_group_filters,
            response.result['endpoint_group']['filters'],
        )
        self.assertEqual(
            endpoint_group_name, response.result['endpoint_group']['name']
        )
        self.assertThat(
            response.result['endpoint_group']['links']['self'],
            matchers.EndsWith(url),
        )

    def test_get_invalid_endpoint_group(self):
        """GET /OS-EP-FILTER/endpoint_groups/{endpoint_group}.

        Invalid endpoint group test case.

        """
        endpoint_group_id = 'foobar'
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.get(url, expected_status=http.client.NOT_FOUND)

    def test_check_endpoint_group(self):
        """HEAD /OS-EP-FILTER/endpoint_groups/{endpoint_group_id}.

        Valid endpoint_group_id test case.

        """
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.head(url, expected_status=http.client.OK)

    def test_check_invalid_endpoint_group(self):
        """HEAD /OS-EP-FILTER/endpoint_groups/{endpoint_group_id}.

        Invalid endpoint_group_id test case.

        """
        endpoint_group_id = 'foobar'
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_patch_endpoint_group(self):
        """PATCH /OS-EP-FILTER/endpoint_groups/{endpoint_group}.

        Valid endpoint group patch test case.

        """
        body = copy.deepcopy(self.DEFAULT_ENDPOINT_GROUP_BODY)
        body['endpoint_group']['filters'] = {'region_id': 'UK'}
        body['endpoint_group']['name'] = 'patch_test'
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        r = self.patch(url, body=body)
        self.assertEqual(endpoint_group_id, r.result['endpoint_group']['id'])
        self.assertEqual(
            body['endpoint_group']['filters'],
            r.result['endpoint_group']['filters'],
        )
        self.assertThat(
            r.result['endpoint_group']['links']['self'], matchers.EndsWith(url)
        )

    def test_patch_nonexistent_endpoint_group(self):
        """PATCH /OS-EP-FILTER/endpoint_groups/{endpoint_group}.

        Invalid endpoint group patch test case.

        """
        body = {'endpoint_group': {'name': 'patch_test'}}
        url = '/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'.format(
            endpoint_group_id='ABC'
        )
        self.patch(url, body=body, expected_status=http.client.NOT_FOUND)

    def test_patch_invalid_endpoint_group(self):
        """PATCH /OS-EP-FILTER/endpoint_groups/{endpoint_group}.

        Valid endpoint group patch test case.

        """
        body = {
            'endpoint_group': {
                'description': 'endpoint group description',
                'filters': {'region': 'UK'},
                'name': 'patch_test',
            }
        }
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.patch(url, body=body, expected_status=http.client.BAD_REQUEST)

        # Perform a GET call to ensure that the content remains
        # the same (as DEFAULT_ENDPOINT_GROUP_BODY) after attempting to update
        # with an invalid filter
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        r = self.get(url)
        del r.result['endpoint_group']['id']
        del r.result['endpoint_group']['links']
        self.assertDictEqual(self.DEFAULT_ENDPOINT_GROUP_BODY, r.result)

    def test_delete_endpoint_group(self):
        """GET /OS-EP-FILTER/endpoint_groups/{endpoint_group}.

        Valid endpoint group test case.

        """
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.delete(url)
        self.get(url, expected_status=http.client.NOT_FOUND)

    def test_delete_invalid_endpoint_group(self):
        """GET /OS-EP-FILTER/endpoint_groups/{endpoint_group}.

        Invalid endpoint group test case.

        """
        endpoint_group_id = 'foobar'
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.delete(url, expected_status=http.client.NOT_FOUND)

    def test_add_endpoint_group_to_project(self):
        """Create a valid endpoint group and project association."""
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )
        self._create_endpoint_group_project_association(
            endpoint_group_id, self.project_id
        )

    def test_add_endpoint_group_to_project_with_invalid_project_id(self):
        """Create an invalid endpoint group and project association."""
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # associate endpoint group with project
        project_id = uuid.uuid4().hex
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, project_id
        )
        self.put(url, expected_status=http.client.NOT_FOUND)

    def test_get_endpoint_group_in_project(self):
        """Test retrieving project endpoint group association."""
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # associate endpoint group with project
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, self.project_id
        )
        self.put(url)
        response = self.get(url)
        self.assertEqual(
            endpoint_group_id,
            response.result['project_endpoint_group']['endpoint_group_id'],
        )
        self.assertEqual(
            self.project_id,
            response.result['project_endpoint_group']['project_id'],
        )

    def test_get_invalid_endpoint_group_in_project(self):
        """Test retrieving project endpoint group association."""
        endpoint_group_id = uuid.uuid4().hex
        project_id = uuid.uuid4().hex
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, project_id
        )
        self.get(url, expected_status=http.client.NOT_FOUND)

    def test_list_endpoint_groups_in_project(self):
        """GET & HEAD /OS-EP-FILTER/projects/{project_id}/endpoint_groups."""
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # associate endpoint group with project
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, self.project_id
        )
        self.put(url)

        url = f'/OS-EP-FILTER/projects/{self.project_id}/endpoint_groups'
        response = self.get(url, expected_status=http.client.OK)

        self.assertEqual(
            endpoint_group_id, response.result['endpoint_groups'][0]['id']
        )

        self.head(url, expected_status=http.client.OK)

    def test_list_endpoint_groups_in_invalid_project(self):
        """Test retrieving from invalid project."""
        project_id = uuid.uuid4().hex
        url = f'/OS-EP-FILTER/projects/{project_id}/endpoint_groups'
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_empty_endpoint_groups_in_project(self):
        """Test when no endpoint groups associated with the project."""
        url = f'/OS-EP-FILTER/projects/{self.project_id}/endpoint_groups'
        response = self.get(url, expected_status=http.client.OK)

        self.assertEqual(0, len(response.result['endpoint_groups']))

        self.head(url, expected_status=http.client.OK)

    def test_check_endpoint_group_to_project(self):
        """Test HEAD with a valid endpoint group and project association."""
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )
        self._create_endpoint_group_project_association(
            endpoint_group_id, self.project_id
        )
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, self.project_id
        )
        self.head(url, expected_status=http.client.OK)

    def test_check_endpoint_group_to_project_with_invalid_project_id(self):
        """Test HEAD with an invalid endpoint group and project association."""
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # create an endpoint group to project association
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, self.project_id
        )
        self.put(url)

        # send a head request with an invalid project id
        project_id = uuid.uuid4().hex
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, project_id
        )
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_list_endpoint_groups(self):
        """GET & HEAD /OS-EP-FILTER/endpoint_groups."""
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # recover all endpoint groups
        url = '/OS-EP-FILTER/endpoint_groups'
        r = self.get(url, expected_status=http.client.OK)
        self.assertNotEmpty(r.result['endpoint_groups'])
        self.assertEqual(
            endpoint_group_id, r.result['endpoint_groups'][0].get('id')
        )

        self.head(url, expected_status=http.client.OK)

    def test_list_endpoint_groups_by_name(self):
        """GET & HEAD /OS-EP-FILTER/endpoint_groups."""
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # retrieve the single endpointgroup by name
        url = '/OS-EP-FILTER/endpoint_groups?name={name}'.format(
            name='endpoint_group_name'
        )
        r = self.get(url, expected_status=http.client.OK)
        self.assertNotEmpty(r.result['endpoint_groups'])
        self.assertEqual(1, len(r.result['endpoint_groups']))
        self.assertEqual(
            endpoint_group_id, r.result['endpoint_groups'][0].get('id')
        )

        self.head(url, expected_status=http.client.OK)

        # try to retrieve a non existant one
        url = '/OS-EP-FILTER/endpoint_groups?name={name}'.format(name='fake')
        r = self.get(url, expected_status=http.client.OK)
        self.assertEqual(0, len(r.result['endpoint_groups']))

    def test_list_projects_associated_with_endpoint_group(self):
        """GET & HEAD /OS-EP-FILTER/endpoint_groups/{endpoint_group}/projects.

        Valid endpoint group test case.

        """
        # create an endpoint group to work with
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # associate endpoint group with project
        self._create_endpoint_group_project_association(
            endpoint_group_id, self.project_id
        )

        # recover list of projects associated with endpoint group
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}' '/projects'
        self.get(url, expected_status=http.client.OK)
        self.head(url, expected_status=http.client.OK)

    def test_list_endpoints_associated_with_endpoint_group(self):
        """GET & HEAD /OS-EP-FILTER/endpoint_groups/{endpoint_group}/endpoints.

        Valid endpoint group test case.

        """
        # create a service
        service_ref = unit.new_service_ref()
        response = self.post('/services', body={'service': service_ref})

        service_id = response.result['service']['id']

        # create an endpoint
        endpoint_ref = unit.new_endpoint_ref(
            service_id=service_id, interface='public', region_id=self.region_id
        )
        response = self.post('/endpoints', body={'endpoint': endpoint_ref})
        endpoint_id = response.result['endpoint']['id']

        # create an endpoint group
        body = copy.deepcopy(self.DEFAULT_ENDPOINT_GROUP_BODY)
        body['endpoint_group']['filters'] = {'service_id': service_id}
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, body
        )

        # create association
        self._create_endpoint_group_project_association(
            endpoint_group_id, self.project_id
        )

        # recover list of endpoints associated with endpoint group
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}' '/endpoints'
        r = self.get(url, expected_status=http.client.OK)
        self.assertNotEmpty(r.result['endpoints'])
        self.assertEqual(endpoint_id, r.result['endpoints'][0].get('id'))
        self.head(url, expected_status=http.client.OK)

    def test_list_endpoints_associated_with_project_endpoint_group(self):
        """GET & HEAD /OS-EP-FILTER/projects/{project_id}/endpoints.

        Valid project, endpoint id, and endpoint group test case.

        """
        # create a temporary service
        service_ref = unit.new_service_ref()
        response = self.post('/services', body={'service': service_ref})
        service_id2 = response.result['service']['id']

        # create additional endpoints
        self._create_endpoint_and_associations(
            self.default_domain_project_id, service_id2
        )
        self._create_endpoint_and_associations(self.default_domain_project_id)

        # create project and endpoint association with default endpoint:
        self.put(self.default_request_url)

        # create an endpoint group that contains a different endpoint
        body = copy.deepcopy(self.DEFAULT_ENDPOINT_GROUP_BODY)
        body['endpoint_group']['filters'] = {'service_id': service_id2}
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, body
        )

        # associate endpoint group with project
        self._create_endpoint_group_project_association(
            endpoint_group_id, self.default_domain_project_id
        )

        # Now get a list of the filtered endpoints
        endpoints_url = f'/OS-EP-FILTER/projects/{self.default_domain_project_id}/endpoints'
        r = self.get(endpoints_url, expected_status=http.client.OK)
        endpoints = self.assertValidEndpointListResponse(r)
        self.assertEqual(2, len(endpoints))
        self.head(endpoints_url, expected_status=http.client.OK)

        # Ensure catalog includes the endpoints from endpoint_group project
        # association, this is needed when a project scoped token is issued
        # and "endpoint_filter.sql" backend driver is in place.
        user_id = uuid.uuid4().hex
        catalog_list = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )
        self.assertEqual(2, len(catalog_list))

        # Now remove project endpoint group association
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, self.default_domain_project_id
        )
        self.delete(url)

        # Now remove endpoint group
        url = f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
        self.delete(url)

        r = self.get(endpoints_url)
        endpoints = self.assertValidEndpointListResponse(r)
        self.assertEqual(1, len(endpoints))

        catalog_list = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )
        self.assertEqual(1, len(catalog_list))

    def test_endpoint_group_project_cleanup_with_project(self):
        # create endpoint group
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # create new project and associate with endpoint_group
        project_ref = unit.new_project_ref(domain_id=self.domain_id)
        r = self.post('/projects', body={'project': project_ref})
        project = self.assertValidProjectResponse(r, project_ref)
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, project['id']
        )
        self.put(url)

        # check that we can recover the project endpoint group association
        self.get(url, expected_status=http.client.OK)
        self.get(url, expected_status=http.client.OK)

        # Now delete the project and then try and retrieve the project
        # endpoint group association again
        self.delete('/projects/{project_id}'.format(project_id=project['id']))
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_endpoint_group_project_cleanup_with_endpoint_group(self):
        # create endpoint group
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # create new project and associate with endpoint_group
        project_ref = unit.new_project_ref(domain_id=self.domain_id)
        r = self.post('/projects', body={'project': project_ref})
        project = self.assertValidProjectResponse(r, project_ref)
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, project['id']
        )
        self.put(url)

        # check that we can recover the project endpoint group association
        self.get(url)

        # now remove the project endpoint group association
        self.delete(url)
        self.get(url, expected_status=http.client.NOT_FOUND)

    def test_removing_an_endpoint_group_project(self):
        # create an endpoint group
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # create an endpoint_group project
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, self.default_domain_project_id
        )
        self.put(url)

        # remove the endpoint group project
        self.delete(url)
        self.get(url, expected_status=http.client.NOT_FOUND)

    def test_remove_endpoint_group_with_project_association(self):
        # create an endpoint group
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # create an endpoint_group project
        project_endpoint_group_url = self._get_project_endpoint_group_url(
            endpoint_group_id, self.default_domain_project_id
        )
        self.put(project_endpoint_group_url)

        # remove endpoint group, the associated endpoint_group project will
        # be removed as well.
        endpoint_group_url = (
            '/OS-EP-FILTER/endpoint_groups/' f'{endpoint_group_id}'
        )
        self.delete(endpoint_group_url)
        self.get(endpoint_group_url, expected_status=http.client.NOT_FOUND)
        self.get(
            project_endpoint_group_url, expected_status=http.client.NOT_FOUND
        )

    @unit.skip_if_cache_disabled('catalog')
    def test_add_endpoint_group_to_project_invalidates_catalog_cache(self):
        # create another endpoint with 'admin' interface which matches
        # 'filters' definition in endpoint group, then there should be two
        # endpoints returned when retrieving v3 catalog if cache works as
        # expected.
        # this should be done at first since `create_endpoint` will also
        # invalidate cache.
        endpoint_id2 = uuid.uuid4().hex
        endpoint2 = unit.new_endpoint_ref(
            service_id=self.service_id,
            region_id=self.region_id,
            interface='admin',
            id=endpoint_id2,
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_id2, endpoint2)

        # create a project and endpoint association.
        self.put(self.default_request_url)

        # there is only one endpoint associated with the default project.
        user_id = uuid.uuid4().hex
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertThat(catalog[0]['endpoints'], matchers.HasLength(1))

        # create an endpoint group.
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # add the endpoint group to default project, bypassing
        # catalog_api API manager.
        PROVIDERS.catalog_api.driver.add_endpoint_group_to_project(
            endpoint_group_id, self.default_domain_project_id
        )

        # can get back only one endpoint from the cache, since the catalog
        # is pulled out from cache.
        invalid_catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertThat(invalid_catalog[0]['endpoints'], matchers.HasLength(1))
        self.assertEqual(catalog, invalid_catalog)

        # remove the endpoint group from default project, and add it again via
        # catalog_api API manager.
        PROVIDERS.catalog_api.driver.remove_endpoint_group_from_project(
            endpoint_group_id, self.default_domain_project_id
        )

        # add the endpoint group to default project.
        PROVIDERS.catalog_api.add_endpoint_group_to_project(
            endpoint_group_id, self.default_domain_project_id
        )

        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        # now, it will return 2 endpoints since the cache has been
        # invalidated.
        self.assertThat(catalog[0]['endpoints'], matchers.HasLength(2))

        ep_id_list = [
            catalog[0]['endpoints'][0]['id'],
            catalog[0]['endpoints'][1]['id'],
        ]
        self.assertCountEqual([self.endpoint_id, endpoint_id2], ep_id_list)

    @unit.skip_if_cache_disabled('catalog')
    def test_remove_endpoint_group_from_project_invalidates_cache(self):
        # create another endpoint with 'admin' interface which matches
        # 'filters' definition in endpoint group, then there should be two
        # endpoints returned when retrieving v3 catalog. But only one
        # endpoint will return after the endpoint group's deletion if cache
        # works as expected.
        # this should be done at first since `create_endpoint` will also
        # invalidate cache.
        endpoint_id2 = uuid.uuid4().hex
        endpoint2 = unit.new_endpoint_ref(
            service_id=self.service_id,
            region_id=self.region_id,
            interface='admin',
            id=endpoint_id2,
        )
        PROVIDERS.catalog_api.create_endpoint(endpoint_id2, endpoint2)

        # create project and endpoint association.
        self.put(self.default_request_url)

        # create an endpoint group.
        endpoint_group_id = self._create_valid_endpoint_group(
            self.DEFAULT_ENDPOINT_GROUP_URL, self.DEFAULT_ENDPOINT_GROUP_BODY
        )

        # add the endpoint group to default project.
        PROVIDERS.catalog_api.add_endpoint_group_to_project(
            endpoint_group_id, self.default_domain_project_id
        )

        # should get back two endpoints, one from endpoint project
        # association, the other one is from endpoint_group project
        # association.
        user_id = uuid.uuid4().hex
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertThat(catalog[0]['endpoints'], matchers.HasLength(2))

        ep_id_list = [
            catalog[0]['endpoints'][0]['id'],
            catalog[0]['endpoints'][1]['id'],
        ]
        self.assertCountEqual([self.endpoint_id, endpoint_id2], ep_id_list)

        # remove endpoint_group project association, bypassing
        # catalog_api API manager.
        PROVIDERS.catalog_api.driver.remove_endpoint_group_from_project(
            endpoint_group_id, self.default_domain_project_id
        )

        # still get back two endpoints, since the catalog is pulled out
        # from cache and the cache haven't been invalidated.
        invalid_catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertThat(invalid_catalog[0]['endpoints'], matchers.HasLength(2))
        self.assertEqual(catalog, invalid_catalog)

        # add back the endpoint_group project association and remove it from
        # manager.
        PROVIDERS.catalog_api.driver.add_endpoint_group_to_project(
            endpoint_group_id, self.default_domain_project_id
        )

        PROVIDERS.catalog_api.remove_endpoint_group_from_project(
            endpoint_group_id, self.default_domain_project_id
        )

        # should only get back one endpoint since the cache has been
        # invalidated after the endpoint_group project association was
        # removed.
        catalog = PROVIDERS.catalog_api.get_v3_catalog(
            user_id, self.default_domain_project_id
        )

        self.assertThat(catalog[0]['endpoints'], matchers.HasLength(1))
        self.assertEqual(self.endpoint_id, catalog[0]['endpoints'][0]['id'])

    def _create_valid_endpoint_group(self, url, body):
        r = self.post(url, body=body)
        return r.result['endpoint_group']['id']

    def _create_endpoint_group_project_association(
        self, endpoint_group_id, project_id
    ):
        url = self._get_project_endpoint_group_url(
            endpoint_group_id, project_id
        )
        self.put(url)

    def _get_project_endpoint_group_url(self, endpoint_group_id, project_id):
        return (
            f'/OS-EP-FILTER/endpoint_groups/{endpoint_group_id}'
            f'/projects/{project_id}'
        )

    def _create_endpoint_and_associations(self, project_id, service_id=None):
        """Create an endpoint associated with service and project."""
        if not service_id:
            # create a new service
            service_ref = unit.new_service_ref()
            response = self.post('/services', body={'service': service_ref})
            service_id = response.result['service']['id']

        # create endpoint
        endpoint_ref = unit.new_endpoint_ref(
            service_id=service_id, interface='public', region_id=self.region_id
        )
        response = self.post('/endpoints', body={'endpoint': endpoint_ref})
        endpoint = response.result['endpoint']

        # now add endpoint to project
        self.put(
            '/OS-EP-FILTER/projects/{project_id}'
            '/endpoints/{endpoint_id}'.format(
                project_id=self.project['id'], endpoint_id=endpoint['id']
            )
        )
        return endpoint



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the credential backend driver in the `keystone.credential`
namespace. Keystone only provides a `sql` driver, so there's no reason to
change this unless you are providing a custom entry point.
"""
    ),
)

provider = cfg.StrOpt(
    'provider',
    default='fernet',
    help=utils.fmt(
        """
Entry point for credential encryption and decryption operations in the
`keystone.credential.provider` namespace. Keystone only provides a `fernet`
driver, so there's no reason to change this unless you are providing a custom
entry point to encrypt and decrypt credentials.
"""
    ),
)

key_repository = cfg.StrOpt(
    'key_repository',
    default='/etc/keystone/credential-keys/',
    help=utils.fmt(
        """
Directory containing Fernet keys used to encrypt and decrypt credentials stored
in the credential backend. Fernet keys used to encrypt credentials have no
relationship to Fernet keys used to encrypt Fernet tokens. Both sets of keys
should be managed separately and require different rotation policies. Do not
share this repository with the repository used to manage keys for Fernet
tokens.
"""
    ),
)

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for caching only on retrieval of user credentials. This has no effect
unless global caching is enabled.
"""
    ),
)

cache_time = cfg.IntOpt(
    'cache_time',
    help=utils.fmt(
        """
Time to cache credential data in seconds. This has no effect unless global
caching is enabled.
"""
    ),
)

auth_ttl = cfg.IntOpt(
    'auth_ttl',
    default=15,
    help=utils.fmt(
        """
The length of time in minutes for which a signed EC2 or S3 token request is
valid from the timestamp contained in the token request.
"""
    ),
)

user_limit = cfg.IntOpt(
    'user_limit',
    default=-1,
    help=utils.fmt(
        """
Maximum number of credentials a user is permitted to create. A value of
-1 means unlimited. If a limit is not set, users are permitted to create
credentials at will, which could lead to bloat in the keystone database
or open keystone to a DoS attack.
"""
    ),
)

GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    driver,
    provider,
    key_repository,
    caching,
    cache_time,
    auth_ttl,
    user_limit,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import http.client
import uuid

from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit import test_v3

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class DomainConfigTestCase(test_v3.RestfulTestCase):
    """Test domain config support."""

    def setUp(self):
        super().setUp()

        self.domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(self.domain['id'], self.domain)
        self.config = {
            'ldap': {
                'url': uuid.uuid4().hex,
                'user_tree_dn': uuid.uuid4().hex,
            },
            'identity': {'driver': uuid.uuid4().hex},
        }

    def test_create_config(self):
        """Call ``PUT /domains/{domain_id}/config``."""
        url = '/domains/{domain_id}/config'.format(domain_id=self.domain['id'])
        r = self.put(
            url,
            body={'config': self.config},
            expected_status=http.client.CREATED,
        )
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        self.assertEqual(self.config, r.result['config'])
        self.assertEqual(self.config, res)

    def test_create_config_invalid_domain(self):
        """Call ``PUT /domains/{domain_id}/config``.

        While creating Identity API-based domain config with an invalid domain
        id provided, the request shall be rejected with a response, 404 domain
        not found.
        """
        invalid_domain_id = uuid.uuid4().hex
        url = f'/domains/{invalid_domain_id}/config'
        self.put(
            url,
            body={'config': self.config},
            expected_status=exception.DomainNotFound.code,
        )

    def test_create_config_twice(self):
        """Check multiple creates don't throw error."""
        self.put(
            '/domains/{domain_id}/config'.format(domain_id=self.domain['id']),
            body={'config': self.config},
            expected_status=http.client.CREATED,
        )
        self.put(
            '/domains/{domain_id}/config'.format(domain_id=self.domain['id']),
            body={'config': self.config},
            expected_status=http.client.OK,
        )

    def test_delete_config(self):
        """Call ``DELETE /domains{domain_id}/config``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        self.delete(
            '/domains/{domain_id}/config'.format(domain_id=self.domain['id'])
        )
        self.get(
            '/domains/{domain_id}/config'.format(domain_id=self.domain['id']),
            expected_status=exception.DomainConfigNotFound.code,
        )

    def test_delete_config_invalid_domain(self):
        """Call ``DELETE /domains{domain_id}/config``.

        While deleting Identity API-based domain config with an invalid domain
        id provided, the request shall be rejected with a response, 404 domain
        not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        invalid_domain_id = uuid.uuid4().hex
        self.delete(
            f'/domains/{invalid_domain_id}/config',
            expected_status=exception.DomainNotFound.code,
        )

    def test_delete_config_by_group(self):
        """Call ``DELETE /domains{domain_id}/config/{group}``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        self.delete(
            '/domains/{domain_id}/config/ldap'.format(
                domain_id=self.domain['id']
            )
        )
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        self.assertNotIn('ldap', res)

    def test_delete_config_by_group_invalid_domain(self):
        """Call ``DELETE /domains{domain_id}/config/{group}``.

        While deleting Identity API-based domain config by group with an
        invalid domain id provided, the request shall be rejected with a
        response 404 domain not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        invalid_domain_id = uuid.uuid4().hex
        self.delete(
            f'/domains/{invalid_domain_id}/config/ldap',
            expected_status=exception.DomainNotFound.code,
        )

    def test_get_head_config(self):
        """Call ``GET & HEAD for /domains{domain_id}/config``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        url = '/domains/{domain_id}/config'.format(domain_id=self.domain['id'])
        r = self.get(url)
        self.assertEqual(self.config, r.result['config'])
        self.head(url, expected_status=http.client.OK)

    def test_get_head_config_by_group(self):
        """Call ``GET & HEAD /domains{domain_id}/config/{group}``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        url = '/domains/{domain_id}/config/ldap'.format(
            domain_id=self.domain['id']
        )
        r = self.get(url)
        self.assertEqual({'ldap': self.config['ldap']}, r.result['config'])
        self.head(url, expected_status=http.client.OK)

    def test_get_head_config_by_group_invalid_domain(self):
        """Call ``GET & HEAD /domains{domain_id}/config/{group}``.

        While retrieving Identity API-based domain config by group with an
        invalid domain id provided, the request shall be rejected with a
        response 404 domain not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        invalid_domain_id = uuid.uuid4().hex
        url = f'/domains/{invalid_domain_id}/config/ldap'
        self.get(url, expected_status=exception.DomainNotFound.code)
        self.head(url, expected_status=exception.DomainNotFound.code)

    def test_get_head_config_by_option(self):
        """Call ``GET & HEAD /domains{domain_id}/config/{group}/{option}``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        url = '/domains/{domain_id}/config/ldap/url'.format(
            domain_id=self.domain['id']
        )
        r = self.get(url)
        self.assertEqual(
            {'url': self.config['ldap']['url']}, r.result['config']
        )
        self.head(url, expected_status=http.client.OK)

    def test_get_head_config_by_option_invalid_domain(self):
        """Call ``GET & HEAD /domains{domain_id}/config/{group}/{option}``.

        While retrieving Identity API-based domain config by option with an
        invalid domain id provided, the request shall be rejected with a
        response 404 domain not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        invalid_domain_id = uuid.uuid4().hex
        url = f'/domains/{invalid_domain_id}/config/ldap/url'
        self.get(url, expected_status=exception.DomainNotFound.code)
        self.head(url, expected_status=exception.DomainNotFound.code)

    def test_get_head_non_existant_config(self):
        """Call ``GET /domains{domain_id}/config when no config defined``."""
        url = '/domains/{domain_id}/config'.format(domain_id=self.domain['id'])
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_get_head_non_existant_config_invalid_domain(self):
        """Call ``GET & HEAD /domains/{domain_id}/config with invalid domain``.

        While retrieving non-existent Identity API-based domain config with an
        invalid domain id provided, the request shall be rejected with a
        response 404 domain not found.
        """
        invalid_domain_id = uuid.uuid4().hex
        url = f'/domains/{invalid_domain_id}/config'
        self.get(url, expected_status=exception.DomainNotFound.code)
        self.head(url, expected_status=exception.DomainNotFound.code)

    def test_get_head_non_existant_config_group(self):
        """Call ``GET /domains/{domain_id}/config/{group_not_exist}``."""
        config = {'ldap': {'url': uuid.uuid4().hex}}
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        url = '/domains/{domain_id}/config/identity'.format(
            domain_id=self.domain['id']
        )
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_get_head_non_existant_config_group_invalid_domain(self):
        """Call ``GET & HEAD /domains/{domain_id}/config/{group}``.

        While retrieving non-existent Identity API-based domain config group
        with an invalid domain id provided, the request shall be rejected with
        a response, 404 domain not found.
        """
        config = {'ldap': {'url': uuid.uuid4().hex}}
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        invalid_domain_id = uuid.uuid4().hex
        url = f'/domains/{invalid_domain_id}/config/identity'
        self.get(url, expected_status=exception.DomainNotFound.code)
        self.head(url, expected_status=exception.DomainNotFound.code)

    def test_get_head_non_existant_config_option(self):
        """Test that Not Found is returned when option doesn't exist.

        Call ``GET & HEAD /domains/{domain_id}/config/{group}/{opt_not_exist}``
        and ensure a Not Found is returned because the option isn't defined
        within the group.
        """
        config = {'ldap': {'url': uuid.uuid4().hex}}
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        url = '/domains/{domain_id}/config/ldap/user_tree_dn'.format(
            domain_id=self.domain['id']
        )
        self.get(url, expected_status=http.client.NOT_FOUND)
        self.head(url, expected_status=http.client.NOT_FOUND)

    def test_get_head_non_existant_config_option_with_invalid_domain(self):
        """Test that Domain Not Found is returned with invalid domain.

        Call ``GET & HEAD /domains/{domain_id}/config/{group}/{opt_not_exist}``

        While retrieving non-existent Identity API-based domain config option
        with an invalid domain id provided, the request shall be rejected with
        a response, 404 domain not found.
        """
        config = {'ldap': {'url': uuid.uuid4().hex}}
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        invalid_domain_id = uuid.uuid4().hex
        url = f'/domains/{invalid_domain_id}/config/ldap/user_tree_dn'
        self.get(url, expected_status=exception.DomainNotFound.code)
        self.head(url, expected_status=exception.DomainNotFound.code)

    def test_update_config(self):
        """Call ``PATCH /domains/{domain_id}/config``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        new_config = {
            'ldap': {'url': uuid.uuid4().hex},
            'identity': {'driver': uuid.uuid4().hex},
        }
        r = self.patch(
            '/domains/{domain_id}/config'.format(domain_id=self.domain['id']),
            body={'config': new_config},
        )
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        expected_config = copy.deepcopy(self.config)
        expected_config['ldap']['url'] = new_config['ldap']['url']
        expected_config['identity']['driver'] = new_config['identity'][
            'driver'
        ]
        self.assertEqual(expected_config, r.result['config'])
        self.assertEqual(expected_config, res)

    def test_update_config_invalid_domain(self):
        """Call ``PATCH /domains/{domain_id}/config``.

        While updating Identity API-based domain config with an invalid domain
        id provided, the request shall be rejected with a response, 404 domain
        not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        new_config = {
            'ldap': {'url': uuid.uuid4().hex},
            'identity': {'driver': uuid.uuid4().hex},
        }
        invalid_domain_id = uuid.uuid4().hex
        self.patch(
            f'/domains/{invalid_domain_id}/config',
            body={'config': new_config},
            expected_status=exception.DomainNotFound.code,
        )

    def test_update_config_group(self):
        """Call ``PATCH /domains/{domain_id}/config/{group}``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        new_config = {
            'ldap': {'url': uuid.uuid4().hex, 'user_filter': uuid.uuid4().hex}
        }
        r = self.patch(
            '/domains/{domain_id}/config/ldap'.format(
                domain_id=self.domain['id']
            ),
            body={'config': new_config},
        )
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        expected_config = copy.deepcopy(self.config)
        expected_config['ldap']['url'] = new_config['ldap']['url']
        expected_config['ldap']['user_filter'] = new_config['ldap'][
            'user_filter'
        ]
        self.assertEqual(expected_config, r.result['config'])
        self.assertEqual(expected_config, res)

    def test_update_config_group_invalid_domain(self):
        """Call ``PATCH /domains/{domain_id}/config/{group}``.

        While updating Identity API-based domain config group with an invalid
        domain id provided, the request shall be rejected with a response,
        404 domain not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        new_config = {
            'ldap': {'url': uuid.uuid4().hex, 'user_filter': uuid.uuid4().hex}
        }
        invalid_domain_id = uuid.uuid4().hex
        self.patch(
            f'/domains/{invalid_domain_id}/config/ldap',
            body={'config': new_config},
            expected_status=exception.DomainNotFound.code,
        )

    def test_update_config_invalid_group(self):
        """Call ``PATCH /domains/{domain_id}/config/{invalid_group}``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )

        # Trying to update a group that is neither whitelisted or sensitive
        # should result in Forbidden.
        invalid_group = uuid.uuid4().hex
        new_config = {
            invalid_group: {
                'url': uuid.uuid4().hex,
                'user_filter': uuid.uuid4().hex,
            }
        }
        self.patch(
            '/domains/{domain_id}/config/{invalid_group}'.format(
                domain_id=self.domain['id'], invalid_group=invalid_group
            ),
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
        )
        # Trying to update a valid group, but one that is not in the current
        # config should result in NotFound
        config = {'ldap': {'suffix': uuid.uuid4().hex}}
        PROVIDERS.domain_config_api.create_config(self.domain['id'], config)
        new_config = {'identity': {'driver': uuid.uuid4().hex}}
        self.patch(
            '/domains/{domain_id}/config/identity'.format(
                domain_id=self.domain['id']
            ),
            body={'config': new_config},
            expected_status=http.client.NOT_FOUND,
        )

    def test_update_config_invalid_group_invalid_domain(self):
        """Call ``PATCH /domains/{domain_id}/config/{invalid_group}``.

        While updating Identity API-based domain config with an invalid group
        and an invalid domain id provided, the request shall be rejected
        with a response, 404 domain not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        invalid_group = uuid.uuid4().hex
        new_config = {
            invalid_group: {
                'url': uuid.uuid4().hex,
                'user_filter': uuid.uuid4().hex,
            }
        }
        invalid_domain_id = uuid.uuid4().hex
        self.patch(
            f'/domains/{invalid_domain_id}/config/{invalid_group}',
            body={'config': new_config},
            expected_status=exception.DomainNotFound.code,
        )

    def test_update_config_option(self):
        """Call ``PATCH /domains/{domain_id}/config/{group}/{option}``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        new_config = {'url': uuid.uuid4().hex}
        r = self.patch(
            '/domains/{domain_id}/config/ldap/url'.format(
                domain_id=self.domain['id']
            ),
            body={'config': new_config},
        )
        res = PROVIDERS.domain_config_api.get_config(self.domain['id'])
        expected_config = copy.deepcopy(self.config)
        expected_config['ldap']['url'] = new_config['url']
        self.assertEqual(expected_config, r.result['config'])
        self.assertEqual(expected_config, res)

    def test_update_config_option_invalid_domain(self):
        """Call ``PATCH /domains/{domain_id}/config/{group}/{option}``.

        While updating Identity API-based domain config option with an invalid
        domain id provided, the request shall be rejected with a response, 404
        domain not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        new_config = {'url': uuid.uuid4().hex}
        invalid_domain_id = uuid.uuid4().hex
        self.patch(
            f'/domains/{invalid_domain_id}/config/ldap/url',
            body={'config': new_config},
            expected_status=exception.DomainNotFound.code,
        )

    def test_update_config_invalid_option(self):
        """Call ``PATCH /domains/{domain_id}/config/{group}/{invalid}``."""
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        invalid_option = uuid.uuid4().hex
        new_config = {'ldap': {invalid_option: uuid.uuid4().hex}}
        # Trying to update an option that is neither whitelisted or sensitive
        # should result in Forbidden.
        self.patch(
            '/domains/{domain_id}/config/ldap/{invalid_option}'.format(
                domain_id=self.domain['id'], invalid_option=invalid_option
            ),
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
        )
        # Trying to update a valid option, but one that is not in the current
        # config should result in NotFound
        new_config = {'suffix': uuid.uuid4().hex}
        self.patch(
            '/domains/{domain_id}/config/ldap/suffix'.format(
                domain_id=self.domain['id']
            ),
            body={'config': new_config},
            expected_status=http.client.NOT_FOUND,
        )

    def test_update_config_invalid_option_invalid_domain(self):
        """Call ``PATCH /domains/{domain_id}/config/{group}/{invalid}``.

        While updating Identity API-based domain config with an invalid option
        and an invalid domain id provided, the request shall be rejected
        with a response, 404 domain not found.
        """
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        invalid_option = uuid.uuid4().hex
        new_config = {'ldap': {invalid_option: uuid.uuid4().hex}}
        invalid_domain_id = uuid.uuid4().hex
        self.patch(
            f'/domains/{invalid_domain_id}/config/ldap/{invalid_option}',
            body={'config': new_config},
            expected_status=exception.DomainNotFound.code,
        )

    def test_get_head_config_default(self):
        """Call ``GET & HEAD /domains/config/default``."""
        # Create a config that overrides a few of the options so that we can
        # check that only the defaults are returned.
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        url = '/domains/config/default'
        r = self.get(url)
        default_config = r.result['config']
        for group in default_config:
            for option in default_config[group]:
                self.assertEqual(
                    getattr(getattr(CONF, group), option),
                    default_config[group][option],
                )
        self.head(url, expected_status=http.client.OK)

    def test_get_head_config_default_by_group(self):
        """Call ``GET & HEAD /domains/config/{group}/default``."""
        # Create a config that overrides a few of the options so that we can
        # check that only the defaults are returned.
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        url = '/domains/config/ldap/default'
        r = self.get(url)
        default_config = r.result['config']
        for option in default_config['ldap']:
            self.assertEqual(
                getattr(CONF.ldap, option), default_config['ldap'][option]
            )
        self.head(url, expected_status=http.client.OK)

    def test_get_head_config_default_by_option(self):
        """Call ``GET & HEAD /domains/config/{group}/{option}/default``."""
        # Create a config that overrides a few of the options so that we can
        # check that only the defaults are returned.
        PROVIDERS.domain_config_api.create_config(
            self.domain['id'], self.config
        )
        url = '/domains/config/ldap/url/default'
        r = self.get(url)
        default_config = r.result['config']
        self.assertEqual(CONF.ldap.url, default_config['url'])
        self.head(url, expected_status=http.client.OK)

    def test_get_head_config_default_by_invalid_group(self):
        """Call ``GET & HEAD for /domains/config/{bad-group}/default``."""
        # First try a valid group, but one we don't support for domain config
        self.get(
            '/domains/config/resource/default',
            expected_status=http.client.FORBIDDEN,
        )
        self.head(
            '/domains/config/resource/default',
            expected_status=http.client.FORBIDDEN,
        )

        # Now try a totally invalid group
        url = f'/domains/config/{uuid.uuid4().hex}/default'
        self.get(url, expected_status=http.client.FORBIDDEN)
        self.head(url, expected_status=http.client.FORBIDDEN)

    def test_get_head_config_default_for_unsupported_group(self):
        # It should not be possible to expose configuration information for
        # groups that the domain configuration API backlists explicitly. Doing
        # so would be a security vulnerability because it would leak sensitive
        # information over the API.
        self.get(
            '/domains/config/ldap/password/default',
            expected_status=http.client.FORBIDDEN,
        )
        self.head(
            '/domains/config/ldap/password/default',
            expected_status=http.client.FORBIDDEN,
        )

    def test_get_head_config_default_for_invalid_option(self):
        """Returning invalid configuration options is invalid."""
        url = f'/domains/config/ldap/{uuid.uuid4().hex}/default'
        self.get(url, expected_status=http.client.FORBIDDEN)
        self.head(url, expected_status=http.client.FORBIDDEN)


class SecurityRequirementsTestCase(test_v3.RestfulTestCase):
    def setUp(self):
        super().setUp()

        # Create a user in the default domain
        self.non_admin_user = unit.create_user(
            PROVIDERS.identity_api, CONF.identity.default_domain_id
        )

        # Create an admin in the default domain
        self.admin_user = unit.create_user(
            PROVIDERS.identity_api, CONF.identity.default_domain_id
        )

        # Create a project in the default domain and a non-admin role
        self.project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        PROVIDERS.resource_api.create_project(self.project['id'], self.project)
        self.non_admin_role = unit.new_role_ref(name='not_admin')
        PROVIDERS.role_api.create_role(
            self.non_admin_role['id'], self.non_admin_role
        )

        # Give the non-admin user a role on the project
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.non_admin_user['id'], self.project['id'], self.role['id']
        )

        # Give the user the admin role on the project, which is technically
        # `self.role` because RestfulTestCase sets that up for us.
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            self.admin_user['id'], self.project['id'], self.role_id
        )

    def _get_non_admin_token(self):
        non_admin_auth_data = self.build_authentication_request(
            user_id=self.non_admin_user['id'],
            password=self.non_admin_user['password'],
            project_id=self.project['id'],
        )
        return self.get_requested_token(non_admin_auth_data)

    def _get_admin_token(self):
        non_admin_auth_data = self.build_authentication_request(
            user_id=self.admin_user['id'],
            password=self.admin_user['password'],
            project_id=self.project['id'],
        )
        return self.get_requested_token(non_admin_auth_data)

    def test_get_head_security_compliance_config_for_default_domain(self):
        """Ask for all security compliance configuration options.

        Support for enforcing security compliance per domain currently doesn't
        exist. Make sure when we ask for security compliance information, it's
        only for the default domain and that it only returns whitelisted
        options.
        """
        password_regex = uuid.uuid4().hex
        password_regex_description = uuid.uuid4().hex
        self.config_fixture.config(
            group='security_compliance', password_regex=password_regex
        )
        self.config_fixture.config(
            group='security_compliance',
            password_regex_description=password_regex_description,
        )
        expected_response = {
            'security_compliance': {
                'password_regex': password_regex,
                'password_regex_description': password_regex_description,
            }
        }
        url = '/domains/{domain_id}/config/{group}'.format(
            domain_id=CONF.identity.default_domain_id,
            group='security_compliance',
        )

        # Make sure regular users and administrators can get security
        # requirement information.
        regular_response = self.get(url, token=self._get_non_admin_token())
        self.assertEqual(regular_response.result['config'], expected_response)
        admin_response = self.get(url, token=self._get_admin_token())
        self.assertEqual(admin_response.result['config'], expected_response)

        # Ensure HEAD requests behave the same way
        self.head(
            url,
            token=self._get_non_admin_token(),
            expected_status=http.client.OK,
        )
        self.head(
            url, token=self._get_admin_token(), expected_status=http.client.OK
        )

    def test_get_security_compliance_config_for_non_default_domain_fails(self):
        """Getting security compliance opts for other domains should fail.

        Support for enforcing security compliance rules per domain currently
        does not exist, so exposing security compliance information for any
        domain other than the default domain should not be allowed.
        """
        # Create a new domain that is not the default domain
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)

        # Set the security compliance configuration options
        password_regex = uuid.uuid4().hex
        password_regex_description = uuid.uuid4().hex
        self.config_fixture.config(
            group='security_compliance', password_regex=password_regex
        )
        self.config_fixture.config(
            group='security_compliance',
            password_regex_description=password_regex_description,
        )
        url = '/domains/{domain_id}/config/{group}'.format(
            domain_id=domain['id'], group='security_compliance'
        )

        # Make sure regular users and administrators are forbidden from doing
        # this.
        self.get(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.get(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

        # Ensure HEAD requests behave the same way
        self.head(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.head(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_get_non_whitelisted_security_compliance_opt_fails(self):
        """We only support exposing a subset of security compliance options.

        Given that security compliance information is sensitive in nature, we
        should make sure that only the options we want to expose are readable
        via the API.
        """
        # Set a security compliance configuration that isn't whitelisted
        self.config_fixture.config(
            group='security_compliance', lockout_failure_attempts=1
        )
        url = '/domains/{domain_id}/config/{group}/{option}'.format(
            domain_id=CONF.identity.default_domain_id,
            group='security_compliance',
            option='lockout_failure_attempts',
        )

        # Make sure regular users and administrators are unable to ask for
        # sensitive information.
        self.get(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.get(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

        # Ensure HEAD requests behave the same way
        self.head(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.head(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_get_security_compliance_password_regex(self):
        """Ask for the security compliance password regular expression."""
        password_regex = uuid.uuid4().hex
        self.config_fixture.config(
            group='security_compliance', password_regex=password_regex
        )
        group = 'security_compliance'
        option = 'password_regex'
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}/{option}'

        # Make sure regular users and administrators can ask for the
        # password regular expression.
        regular_response = self.get(url, token=self._get_non_admin_token())
        self.assertEqual(
            regular_response.result['config'][option], password_regex
        )
        admin_response = self.get(url, token=self._get_admin_token())
        self.assertEqual(
            admin_response.result['config'][option], password_regex
        )

        # Ensure HEAD requests behave the same way
        self.head(
            url,
            token=self._get_non_admin_token(),
            expected_status=http.client.OK,
        )
        self.head(
            url, token=self._get_admin_token(), expected_status=http.client.OK
        )

    def test_get_security_compliance_password_regex_description(self):
        """Ask for the security compliance password regex description."""
        password_regex_description = uuid.uuid4().hex
        self.config_fixture.config(
            group='security_compliance',
            password_regex_description=password_regex_description,
        )
        group = 'security_compliance'
        option = 'password_regex_description'
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}/{option}'

        # Make sure regular users and administrators can ask for the
        # password regular expression.
        regular_response = self.get(url, token=self._get_non_admin_token())
        self.assertEqual(
            regular_response.result['config'][option],
            password_regex_description,
        )
        admin_response = self.get(url, token=self._get_admin_token())
        self.assertEqual(
            admin_response.result['config'][option], password_regex_description
        )

        # Ensure HEAD requests behave the same way
        self.head(
            url,
            token=self._get_non_admin_token(),
            expected_status=http.client.OK,
        )
        self.head(
            url, token=self._get_admin_token(), expected_status=http.client.OK
        )

    def test_get_security_compliance_password_regex_returns_none(self):
        """When an option isn't set, we should explicitly return None."""
        group = 'security_compliance'
        option = 'password_regex'
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}/{option}'

        # Make sure regular users and administrators can ask for the password
        # regular expression, but since it isn't set the returned value should
        # be None.
        regular_response = self.get(url, token=self._get_non_admin_token())
        self.assertIsNone(regular_response.result['config'][option])
        admin_response = self.get(url, token=self._get_admin_token())
        self.assertIsNone(admin_response.result['config'][option])

        # Ensure HEAD requests behave the same way
        self.head(
            url,
            token=self._get_non_admin_token(),
            expected_status=http.client.OK,
        )
        self.head(
            url, token=self._get_admin_token(), expected_status=http.client.OK
        )

    def test_get_security_compliance_password_regex_desc_returns_none(self):
        """When an option isn't set, we should explicitly return None."""
        group = 'security_compliance'
        option = 'password_regex_description'
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}/{option}'

        # Make sure regular users and administrators can ask for the password
        # regular expression description, but since it isn't set the returned
        # value should be None.
        regular_response = self.get(url, token=self._get_non_admin_token())
        self.assertIsNone(regular_response.result['config'][option])
        admin_response = self.get(url, token=self._get_admin_token())
        self.assertIsNone(admin_response.result['config'][option])

        # Ensure HEAD requests behave the same way
        self.head(
            url,
            token=self._get_non_admin_token(),
            expected_status=http.client.OK,
        )
        self.head(
            url, token=self._get_admin_token(), expected_status=http.client.OK
        )

    def test_get_security_compliance_config_with_user_from_other_domain(self):
        """Make sure users from other domains can access password requirements.

        Even though a user is in a separate domain, they should be able to see
        the security requirements for the deployment. This is because security
        compliance is not yet implemented on a per domain basis. Once that
        happens, then this should no longer be possible since a user should
        only care about the security compliance requirements for the domain
        that they are in.
        """
        # Make a new domain
        domain = unit.new_domain_ref()
        PROVIDERS.resource_api.create_domain(domain['id'], domain)

        # Create a user in the new domain
        user = unit.create_user(PROVIDERS.identity_api, domain['id'])

        # Create a project in the new domain
        project = unit.new_project_ref(domain_id=domain['id'])
        PROVIDERS.resource_api.create_project(project['id'], project)

        # Give the new user a non-admin role on the project
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            user['id'], project['id'], self.non_admin_role['id']
        )

        # Set our security compliance config values, we do this after we've
        # created our test user otherwise password validation will fail with a
        # uuid type regex.
        password_regex = uuid.uuid4().hex
        password_regex_description = uuid.uuid4().hex
        group = 'security_compliance'
        self.config_fixture.config(group=group, password_regex=password_regex)
        self.config_fixture.config(
            group=group, password_regex_description=password_regex_description
        )

        # Get a token for the newly created user scoped to the project in the
        # non-default domain and use it to get the password security
        # requirements.
        user_token = self.build_authentication_request(
            user_id=user['id'],
            password=user['password'],
            project_id=project['id'],
        )
        user_token = self.get_requested_token(user_token)
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}'
        response = self.get(url, token=user_token)
        self.assertEqual(
            response.result['config'][group]['password_regex'], password_regex
        )
        self.assertEqual(
            response.result['config'][group]['password_regex_description'],
            password_regex_description,
        )

        # Ensure HEAD requests behave the same way
        self.head(url, token=user_token, expected_status=http.client.OK)

    def test_update_security_compliance_config_group_fails(self):
        """Make sure that updates to the entire security group section fail.

        We should only allow the ability to modify a deployments security
        compliance rules through configuration. Especially since it's only
        enforced on the default domain.
        """
        new_config = {
            'security_compliance': {
                'password_regex': uuid.uuid4().hex,
                'password_regex_description': uuid.uuid4().hex,
            }
        }
        url = '/domains/{domain_id}/config/{group}'.format(
            domain_id=CONF.identity.default_domain_id,
            group='security_compliance',
        )

        # Make sure regular users and administrators aren't allowed to modify
        # security compliance configuration through the API.
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_update_security_compliance_password_regex_fails(self):
        """Make sure any updates to security compliance options fail."""
        group = 'security_compliance'
        option = 'password_regex'
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}/{option}'
        new_config = {group: {option: uuid.uuid4().hex}}

        # Make sure regular users and administrators aren't allowed to modify
        # security compliance configuration through the API.
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_update_security_compliance_password_regex_description_fails(self):
        """Make sure any updates to security compliance options fail."""
        group = 'security_compliance'
        option = 'password_regex_description'
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}/{option}'
        new_config = {group: {option: uuid.uuid4().hex}}

        # Make sure regular users and administrators aren't allowed to modify
        # security compliance configuration through the API.
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_update_non_whitelisted_security_compliance_option_fails(self):
        """Updating security compliance options through the API is not allowed.

        Requests to update anything in the security compliance group through
        the API should be Forbidden. This ensures that we are covering cases
        where the option being updated isn't in the white list.
        """
        group = 'security_compliance'
        option = 'lockout_failure_attempts'
        url = f'/domains/{CONF.identity.default_domain_id}/config/{group}/{option}'
        new_config = {group: {option: 1}}

        # Make sure this behavior is not possible for regular users or
        # administrators.
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.patch(
            url,
            body={'config': new_config},
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_delete_security_compliance_group_fails(self):
        """The security compliance group shouldn't be deleteable."""
        url = '/domains/{domain_id}/config/{group}/'.format(
            domain_id=CONF.identity.default_domain_id,
            group='security_compliance',
        )

        # Make sure regular users and administrators can't delete the security
        # compliance configuration group.
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_delete_security_compliance_password_regex_fails(self):
        """The security compliance options shouldn't be deleteable."""
        url = '/domains/{domain_id}/config/{group}/{option}'.format(
            domain_id=CONF.identity.default_domain_id,
            group='security_compliance',
            option='password_regex',
        )

        # Make sure regular users and administrators can't delete the security
        # compliance configuration group.
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_delete_security_compliance_password_regex_description_fails(self):
        """The security compliance options shouldn't be deleteable."""
        url = '/domains/{domain_id}/config/{group}/{option}'.format(
            domain_id=CONF.identity.default_domain_id,
            group='security_compliance',
            option='password_regex_description',
        )

        # Make sure regular users and administrators can't delete the security
        # compliance configuration group.
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )

    def test_delete_non_whitelisted_security_compliance_options_fails(self):
        """The security compliance options shouldn't be deleteable."""
        url = '/domains/{domain_id}/config/{group}/{option}'.format(
            domain_id=CONF.identity.default_domain_id,
            group='security_compliance',
            option='lockout_failure_attempts',
        )

        # Make sure regular users and administrators can't delete the security
        # compliance configuration group.
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_non_admin_token(),
        )
        self.delete(
            url,
            expected_status=http.client.FORBIDDEN,
            token=self._get_admin_token(),
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The policy API is now aware of system scope and default roles."
)

deprecated_get_policy = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_policy',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_list_policies = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_policies',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_update_policy = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_policy',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_create_policy = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_policy',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)

deprecated_delete_policy = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_policy',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)


policy_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_policy',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        # This API isn't really exposed to usable, it's actually deprecated.
        # More-or-less adding scope_types to be consistent with other policies.
        scope_types=['system', 'project'],
        description='Show policy details.',
        operations=[{'path': '/v3/policies/{policy_id}', 'method': 'GET'}],
        deprecated_rule=deprecated_get_policy,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_policies',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List policies.',
        operations=[{'path': '/v3/policies', 'method': 'GET'}],
        deprecated_rule=deprecated_list_policies,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_policy',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create policy.',
        operations=[{'path': '/v3/policies', 'method': 'POST'}],
        deprecated_rule=deprecated_create_policy,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_policy',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Update policy.',
        operations=[{'path': '/v3/policies/{policy_id}', 'method': 'PATCH'}],
        deprecated_rule=deprecated_update_policy,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_policy',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete policy.',
        operations=[{'path': '/v3/policies/{policy_id}', 'method': 'DELETE'}],
        deprecated_rule=deprecated_delete_policy,
    ),
]


def list_rules():
    return policy_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config
from sqlalchemy import pool

from keystone.common.sql import core
from keystone.common.sql.migrations import autogen

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# interpret the config file for Python logging unless we're told not to;
# this line sets up loggers basically.
if config.attributes.get('configure_logger', True):
    fileConfig(config.config_file_name)

# keystone model MetaData object
target_metadata = core.ModelBase.metadata


def include_object(object, name, type_, reflected, compare_to):
    BORKED_COLUMNS = ()

    BORKED_UNIQUE_CONSTRAINTS = ()

    BORKED_FK_CONSTRAINTS = (
        # removed fks
        ('application_credential_access_rule', ['access_rule_id']),
        ('limit', ['registered_limit_id']),
        ('registered_limit', ['service_id']),
        ('registered_limit', ['region_id']),
        ('endpoint', ['region_id']),
        # added fks
        ('application_credential_access_rule', ['access_rule_id']),
        ('endpoint', ['region_id']),
        ('assignment', ['role_id']),
    )

    BORKED_INDEXES = (
        # removed indexes
        ('access_rule', ['external_id']),
        ('access_rule', ['user_id']),
        ('revocation_event', ['revoked_at']),
        ('system_assignment', ['actor_id']),
        ('user', ['default_project_id']),
        # added indexes
        ('access_rule', ['external_id']),
        ('access_rule', ['user_id']),
        ('access_token', ['consumer_id']),
        ('endpoint', ['service_id']),
        ('revocation_event', ['revoked_at']),
        ('user', ['default_project_id']),
        ('user_group_membership', ['group_id']),
        (
            'trust',
            [
                'trustor_user_id',
                'trustee_user_id',
                'project_id',
                'impersonation',
                'expires_at',
                'expires_at_int',
            ],
        ),
    )

    # NOTE(stephenfin): By skipping these items, we skip *all* changes to the
    # affected item. However, we only want to skip the actual things we know
    # about untl we have enough time to fix them. These issues are listed in
    # keystone.tests.unit.common.sql.test_upgrades.KeystoneModelsMigrationsSync
    # However, this isn't an issue since the test is more specific and will
    # catch other issues and anyone making changes to the columns and hoping to
    # autogenerate them would need to fix the latent issue first anyway.
    if type_ == 'column':
        return (object.table.name, name) not in BORKED_COLUMNS

    if type_ == 'unique_constraint':
        columns = [c.name for c in object.columns]
        return (object.table.name, columns) not in BORKED_UNIQUE_CONSTRAINTS

    if type_ == 'foreign_key_constraint':
        columns = [c.name for c in object.columns]
        return (object.table.name, columns) not in BORKED_FK_CONSTRAINTS

    if type_ == 'index':
        columns = [c.name for c in object.columns]
        return (object.table.name, columns) not in BORKED_INDEXES

    return True


def include_name(name, type_, parent_names):
    """Determine which tables or columns to skip.

    This is used where we have migrations that are out-of-sync with the models.
    """
    REMOVED_TABLES = ('token',)

    if type_ == 'table':
        return name not in REMOVED_TABLES

    return True


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL and not an Engine, though an
    Engine is acceptable here as well.  By skipping the Engine creation we
    don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the script output.
    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(
        url=url,
        target_metadata=target_metadata,
        render_as_batch=True,
        include_name=include_name,
        include_object=include_object,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a connection
    with the context.

    This is modified from the default based on the below, since we want to
    share an engine when unit testing so in-memory database testing actually
    works.

    https://alembic.sqlalchemy.org/en/latest/cookbook.html#connection-sharing
    """
    connectable = config.attributes.get('connection', None)

    if connectable is None:
        # only create Engine if we don't have a Connection from the outside
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )
        with connectable.connect() as connection:
            context.configure(
                connection=connection,
                target_metadata=target_metadata,
                render_as_batch=True,
                include_name=include_name,
                include_object=include_object,
                process_revision_directives=autogen.process_revision_directives,  # noqa: E501
            )

            with context.begin_transaction():
                context.run_migrations()
    else:
        context.configure(
            connection=connectable,
            target_metadata=target_metadata,
            render_as_batch=True,
            include_name=include_name,
            include_object=include_object,
            process_revision_directives=autogen.process_revision_directives,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright 2012 OpenStack Foundation
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 - 2012 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# A couple common constants for Auth data

# Header used to transmit the auth token
AUTH_TOKEN_HEADER = 'X-Auth-Token'  # nosec


# Header used to transmit the auth receipt
AUTH_RECEIPT_HEADER = 'Openstack-Auth-Receipt'


# Header used to transmit the subject token
SUBJECT_TOKEN_HEADER = 'X-Subject-Token'  # nosec

# Environment variable used to convey the Keystone auth context,
# the user credential used for policy enforcement.
AUTH_CONTEXT_ENV = 'KEYSTONE_AUTH_CONTEXT'

# Header set by versions of keystonemiddleware that understand application
# credential access rules
ACCESS_RULES_HEADER = 'OpenStack-Identity-Access-Rules'
